from typing import Dict

from pait.model.status import PaitStatus

__all__ = ["get_color"]
_pait_status_num_dict: Dict[PaitStatus, int] = {i: index for index, i in enumerate(PaitStatus.__members__.values())}


def get_color(pait_status: PaitStatus) -> str:
    pait_status_num: int = _pait_status_num_dict[pait_status]
    if pait_status_num <= _pait_status_num_dict[PaitStatus.dev]:
        return "#DC143C"
    elif pait_status_num <= _pait_status_num_dict[PaitStatus.test]:
        return "#00BFFF"
    elif pait_status_num <= _pait_status_num_dict[PaitStatus.release]:
        return "#32CD32"
    else:
        return "#DC143C"



import inspect
import re
from dataclasses import dataclass
from pathlib import Path
from types import ModuleType
from typing import Any, Dict, List, Optional, Type

from pait.util.grpc_inspect.types import Message


def get_proto_msg_path(line: str, re_str: str) -> str:
    module_path_find_list = re.findall(re_str, line)
    if len(module_path_find_list) != 1:
        raise ValueError("module path not found")
    return module_path_find_list[0]


@dataclass()
class GrpcModel(object):
    invoke_name: str
    method: str
    # func: Callable
    request: Type[Message] = Message
    response: Type[Message] = Message
    desc: str = ""
    service_desc: str = ""


class ParseStub(object):
    def __init__(self, stub: Any, parse_msg_desc: Optional[str] = None):
        self._stub: Any = stub
        self._parse_msg_desc: Optional[str] = parse_msg_desc
        self.name: str = self._stub.__name__
        self._method_dict: Dict[str, GrpcModel] = {}
        self._line_list: List[str] = inspect.getsource(stub).split("\n")

        class_module: Optional[ModuleType] = inspect.getmodule(stub)
        if not class_module:
            raise RuntimeError(f"Can not found {stub} module")
        self._class_module: ModuleType = class_module

        self._filename_desc_dict: Dict[str, Dict[str, Dict[str, str]]] = {}

        # self._gen_rpc_metadate()
        self._parse()

    @property
    def method_dict(self) -> Dict[str, GrpcModel]:
        return self._method_dict

    def _get_desc_from_pyi_file(self, filename: str) -> Dict[str, Dict[str, str]]:
        if filename in self._filename_desc_dict:
            return self._filename_desc_dict[filename]

        with open(filename, "r") as f:
            pyi_content: str = f.read()
        line_list = pyi_content.split("\n")

        _comment_model: bool = False
        _doc: str = ""
        _field_name: str = ""
        message_str: str = ""
        message_field_dict: Dict[str, Dict[str, str]] = {}
        for index, line in enumerate(line_list):
            if "class" in line:
                if line.endswith("google.protobuf.message.Message):"):
                    match_list = re.findall(r"class (.+)\(google.protobuf.message.Message", line)
                    if not match_list:
                        continue
                    message_str = match_list[0]
                else:
                    message_str = ""
                continue

            if message_str:
                line = line.strip()
                if _comment_model:
                    _doc += "\n" + line

                if not _comment_model and line.startswith('"""') and not line_list[index - 1].startswith("class"):
                    # start add doc
                    _field_name = line_list[index - 1].split(":")[0].strip()
                    _comment_model = True
                    _doc = line
                if (line.endswith('"""') or line == '"""') and _comment_model:
                    # end add doc
                    _comment_model = False
                    if message_str not in message_field_dict:
                        message_field_dict[message_str] = {}
                    message_field_dict[message_str][_field_name] = _doc.replace('"""', "")

        self._filename_desc_dict[filename] = message_field_dict
        return message_field_dict

    def _get_desc_from_proto_file(self, filename: str) -> Dict[str, Dict[str, str]]:
        if filename in self._filename_desc_dict:
            return self._filename_desc_dict[filename]

        with open(filename, "r") as f:
            protobuf_content: str = f.read()
        message_stack: List[str] = []
        message_field_dict: Dict[str, Dict[str, str]] = {}
        _field: str = ""
        _doc: str = ""
        _comment_model: bool = False

        for line in protobuf_content.split("\n"):
            _comment_model = False
            line_list: List[str] = line.split()
            for index, column in enumerate(line_list):
                if _comment_model:
                    _doc += column + " "
                    continue

                if column == "message" and line_list[index + 2] == "{":
                    # message parse start
                    message_stack.append(line_list[index + 1])
                    continue
                if message_stack:
                    if column == "}":
                        # message parse end
                        message_stack.pop()
                    elif column == "//":
                        # comment start
                        _comment_model = True
                        _doc += "\n"
                    elif column == "=":
                        # get field name
                        _field = line_list[index - 1]
                    elif column[-1] == ";":
                        # field parse end
                        if _doc:
                            message_str: str = message_stack[-1]
                            if message_str not in message_field_dict:
                                message_field_dict[message_str] = {}

                            message_field_dict[message_str][_field] = _doc
                        _field = ""
                        _doc = ""
        self._filename_desc_dict[filename] = message_field_dict
        return message_field_dict

    def _gen_message(self, line: str, match_str: str) -> Type[Message]:
        module_path: str = get_proto_msg_path(line, match_str)
        module_path_list: List[str] = module_path.split(".")
        message_module: ModuleType = getattr(self._class_module, module_path_list[0])
        message_model: Type[Message] = getattr(message_module, module_path_list[1])

        if not issubclass(message_model, Message):
            raise RuntimeError("Can not found message")

        if not self._parse_msg_desc:
            return message_model
        if Path(self._parse_msg_desc).exists():
            proto_file_name = message_model.DESCRIPTOR.file.name
            if proto_file_name.endswith("empty.proto"):
                return message_model
            file_str: str = self._parse_msg_desc
            if not file_str.endswith("/"):
                file_str += "/"
            message_field_dict: Dict[str, Dict[str, str]] = self._get_desc_from_proto_file(file_str + proto_file_name)
        elif self._parse_msg_desc == "by_mypy":
            pyi_file_name = inspect.getfile(message_module) + "i"
            if pyi_file_name.endswith("empty_pb2.pyi"):
                return message_model
            if not Path(pyi_file_name).exists():
                raise RuntimeError(f"Can not found {message_module} pyi file")
            message_field_dict = self._get_desc_from_pyi_file(pyi_file_name)
        elif self._parse_msg_desc == "by_pait":
            raise NotImplementedError("by_pait")
        else:
            raise ValueError(
                f"parse_msg_desc must be exist path or `by_mypy` or `by_pait`, not {self._parse_msg_desc})"
            )

        if message_model.__name__ in message_field_dict:
            message_model.__field_doc_dict__ = message_field_dict[message_model.__name__]
        return message_model

    def _parse(self) -> None:
        line_list: List[str] = self._line_list

        service_class_name: str = self._stub.__name__.replace("Stub", "Servicer")
        service_class: Type = getattr(self._class_module, service_class_name)

        for index, line in enumerate(line_list):
            if "self." not in line:
                continue
            if "channel.unary_unary" not in line:
                continue

            invoke_name: str = line.split("=")[0].replace("self.", "").strip()
            method: str = line_list[index + 1].strip()[1:-2]
            request: Type[Message] = self._gen_message(
                line_list[index + 2], r"request_serializer=(.+).SerializeToString"
            )
            response: Type[Message] = self._gen_message(line_list[index + 3], r"response_deserializer=(.+).FromString")
            self._method_dict[method] = GrpcModel(
                invoke_name=invoke_name,
                method=method,
                desc=service_class.__dict__[invoke_name].__doc__ or "",
                service_desc=service_class.__doc__ or "",
                request=request,
                response=response,
            )



from typing import Any, Callable, List, Type

from google.protobuf.json_format import MessageToDict  # type: ignore
from google.protobuf.message import Message  # type: ignore
from pydantic import BaseModel
from tornado.web import Application, RequestHandler
from typing_extensions import Self  # type: ignore

from pait.app.base.grpc_route import AsyncGrpcGatewayRoute as BaseGrpcRouter
from pait.app.base.grpc_route import GrpcModel
from pait.app.tornado import pait as tornado_pait


def tornado_make_response(_: Any, resp_dict: dict) -> dict:
    return resp_dict


class GrpcGatewayRoute(BaseGrpcRouter):
    pait = tornado_pait
    make_response = tornado_make_response
    request_handler: Type[RequestHandler] = RequestHandler

    def with_request(self, request: Type[RequestHandler]) -> None:
        self.request_handler = request

    def gen_route(
        self, method_name: str, grpc_model: GrpcModel, request_pydantic_model_class: Type[BaseModel]
    ) -> Callable:
        async def _route(
            route_self: Self,
            request_pydantic_model: request_pydantic_model_class,  # type: ignore
        ) -> None:
            func: Callable = self.get_grpc_func(method_name)
            request_dict: dict = request_pydantic_model.dict()  # type: ignore
            request_msg: Message = self.get_msg_from_dict(grpc_model.request, request_dict)
            grpc_msg: Message = await func(request_msg)
            resp_dict: dict = self._make_response(self.get_dict_from_msg(grpc_msg))
            route_self.write(resp_dict)

        return _route

    def _add_route(self, app: Application) -> Any:
        prefix: str = self.prefix
        if prefix.endswith("/"):
            prefix = prefix[:-1]
        route_list: List = []
        for parse_stub in self.parse_stub_list:
            for method_name, grpc_model in parse_stub.method_dict.items():
                _route, grpc_pait_model = self._gen_route_func(method_name, grpc_model)
                if not _route:
                    continue

                route_class = type(
                    grpc_model.method.split(".")[-1],
                    (self.request_handler,),
                    {"__model__": f"{__name__}.{self.__class__.__name__}.gen_route_func.<locals>"},
                )
                setattr(route_class, "post", _route)

                # grpc http method only POST
                route_list.append((r"{}{}".format(prefix, self.url_handler(grpc_pait_model.url)), route_class))

        app.wildcard_router.add_rules(route_list)
        from tornado.web import AnyMatches, Rule, _ApplicationRouter

        app.default_router = _ApplicationRouter(app, [Rule(AnyMatches(), app.wildcard_router)])



import inspect
import logging
from typing import Any, Optional, Type

from pydantic import BaseModel

from pait.exceptions import TipException
from pait.field import BaseField
from pait.util import FuncSig

_indent: str = 4 * " "

__all__ = ["gen_tip_exc"]


def gen_tip_exc(_object: Any, exception: "Exception", parameter: Optional[inspect.Parameter] = None) -> Exception:
    """Help users understand which parameter is wrong"""
    if _object is None:
        return exception
    if isinstance(exception, TipException):
        return exception
    if parameter:
        param_value: BaseField = parameter.default
        annotation: Type[BaseModel] = parameter.annotation
        param_name: str = parameter.name

        parameter_value_name: str = param_value.__class__.__name__
        if param_value is parameter.empty:
            param_str: str = f"{param_name}: {annotation}"
        else:
            param_str = f"{param_name}: {annotation} = {parameter_value_name}"
            if isinstance(param_value, BaseField):
                param_str += f"({param_value})"
    else:
        param_str = ""

    file: Optional[str] = None
    if isinstance(_object, FuncSig):
        _object = _object.func
    if inspect.isfunction(_object):
        title: str = "def"
        if inspect.iscoroutinefunction(_object):
            title = "async def"
        file = inspect.getfile(_object)
        line: int = inspect.getsourcelines(_object)[1]
        error_object_name: str = _object.__name__
        error_source_tip: str = (
            f"{title} {error_object_name}(\n"
            f"{_indent}...\n"
            f"{_indent}{param_str} <-- error\n"
            f"{_indent}...\n):\n"
            f"{_indent}pass"
        )
    else:
        module: Any = inspect.getmodule(_object)
        if module:
            file = module.__file__

        class_object = _object.__class__ if not inspect.isclass(_object) else _object
        line = inspect.getsourcelines(class_object)[1]
        error_object_name = class_object.__name__
        if "class" in error_object_name:
            error_object_name = str(class_object)  # pragma: no cover
        error_source_tip = f"class: `{error_object_name}`  attributes error\n    {param_str}"

    logging.debug(error_source_tip)
    exc_msg: str = (
        f"{str(exception)} for {_object}   Customer Traceback:\n"
        f'{_indent}File "{file}",'
        f" line {line},"
        f" in {error_object_name}."
    )
    return TipException(exc_msg, exception)



from typing import Any, Optional

from starlette.responses import JSONResponse

from pait.plugin.base import PluginManager, PluginProtocol


class JsonProtocol(PluginProtocol):
    status_code: int
    headers: Optional[dict]
    media_type: Optional[str]

    def gen_response(self, response: Any) -> JSONResponse:
        return JSONResponse(response, status_code=self.status_code, headers=self.headers, media_type=self.media_type)

    @classmethod
    def build(  # type: ignore
        cls,
        status_code: Optional[int] = None,
        headers: Optional[dict] = None,
        media_type: Optional[str] = None,
        **kwargs: Any,
    ) -> "PluginManager":
        return super().build(
            status_code=status_code or 200,
            headers=headers or {},
            media_type=media_type,
            **kwargs,
        )



from typing import Any, Dict, List

from pait.exceptions import CheckValueError
from pait.plugin.base import PluginManager, PluginProtocol
from pait.util import gen_tip_exc

__all__ = ["RequiredPlugin", "AsyncRequiredPlugin"]


class RequiredPlugin(PluginProtocol):
    """
    Check dependencies between parameters
    """

    required_dict: Dict[str, List[str]]

    is_pre_core: bool = False

    def check_param(self, **kwargs: Any) -> None:
        try:
            for pre_param, param_list in self.required_dict.items():
                if pre_param not in kwargs or not kwargs[pre_param]:
                    continue
                for param in param_list:
                    if kwargs.get(param, None) is None:
                        raise CheckValueError(
                            f"{pre_param} requires param {' and '.join(param_list)}, which if not none"
                        )
        except Exception as e:
            raise e from gen_tip_exc(self.pait_core_model.func, e)

    @classmethod
    def build(cls, *, required_dict: Dict[str, List[str]]) -> "PluginManager":  # type: ignore
        return super().build(required_dict=required_dict)

    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        self.check_param(**kwargs)
        return self.call_next(*args, **kwargs)


class AsyncRequiredPlugin(RequiredPlugin):
    """"""



import inspect
import json
import os
import sys
from dataclasses import MISSING
from datetime import date, datetime
from decimal import Decimal
from enum import Enum
from functools import wraps
from json import JSONEncoder
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    ForwardRef,
    List,
    Optional,
    Set,
    Tuple,
    Type,
    TypeVar,
    Union,
    get_type_hints,
)

from pydantic import BaseModel
from pydantic.fields import Undefined, UndefinedType

from pait.field import BaseField, Depends, is_pait_field
from pait.model.template import TemplateVar
from pait.types import ParamSpec

from ._types import ParseTypeError, parse_typing

if TYPE_CHECKING:
    from pait.model.response import PaitBaseResponseModel

__all__ = [
    "gen_example_dict_from_pydantic_base_model",
    "gen_example_dict_from_schema",
    "gen_example_json_from_schema",
    "get_parameter_list_from_pydantic_basemodel",
    "get_parameter_list_from_class",
    "CustomJSONEncoder",
    "http_method_tuple",
    "json_type_default_value_dict",
    "python_type_default_value_dict",
    "get_pait_response_model",
    "example_value_handle",
    "ignore_pre_check",
    "gen_example_value_from_python",
    "get_real_annotation",
    "create_factory",
]
ignore_pre_check: bool = bool(os.environ.get("PAIT_IGNORE_PRE_CHECK", False))
http_method_tuple: Tuple[str, ...] = ("get", "post", "head", "options", "delete", "put", "trace", "patch")

json_type_default_value_dict: Dict[str, Any] = {
    "null": None,
    "bool": True,
    "boolean": True,
    "string": "",
    "number": 0.0,
    "float": 0.0,
    "integer": 0,
    "object": {},
    "array": [],
}

python_type_default_value_dict: Dict[type, Any] = {
    bool: True,
    float: 0.0,
    int: 0,
    str: "",
    list: [],
    tuple: (),
    dict: {},
    datetime: datetime.fromtimestamp(0),
    Decimal: Decimal("0.0"),
}


P = ParamSpec("P")
R_T = TypeVar("R_T")


def create_factory(func: Callable[P, R_T]) -> Callable[P, R_T]:  # type: ignore
    @wraps(func)
    def wrapper(*args: P.args, **kwargs: P.kwargs) -> R_T:  # type: ignore
        return lambda: func(*args, **kwargs)  # type: ignore

    return wrapper


def example_value_handle(example_value: Any) -> Any:
    if getattr(example_value, "__call__", None):
        example_value = example_value()
    elif isinstance(example_value, Enum):
        example_value = example_value.value
    elif isinstance(example_value, TemplateVar):
        example_value = example_value.get_value_from_template_context()
    return example_value


def get_real_annotation(annotation: Union[Type, str], target_obj: Any) -> Type:
    if not isinstance(annotation, str):
        return annotation
    if inspect.isclass(target_obj) and annotation in target_obj.__dict__:
        new_annotation: Type = target_obj.__dict__[annotation]
    else:
        # get real type
        value: ForwardRef = ForwardRef(annotation, is_argument=False)
        new_annotation = value._evaluate(sys.modules[target_obj.__module__].__dict__, None)  # type: ignore
        if not new_annotation:
            raise RuntimeError(f"get real annotation from {target_obj} fail")  # pragma: no cover
    return new_annotation


def get_pydantic_annotation(key: str, pydantic_base_model: Type[BaseModel]) -> Type:
    annotation: Any = MISSING
    for base in reversed(pydantic_base_model.__mro__):
        ann: Union[str, Type] = base.__dict__.get("__annotations__", {}).get(key, MISSING)
        if ann is not MISSING:
            annotation = ann
            break
    if annotation is MISSING:
        raise RuntimeError(f"get annotation from {pydantic_base_model} fail")  # pragma: no cover
    annotation = get_real_annotation(annotation, pydantic_base_model)

    if getattr(annotation, "real", None) and annotation != bool:
        # support like pydantic.ConstrainedIntValue
        annotation = annotation.real.__objclass__  # type: ignore
    return annotation


def get_pait_response_model(
    response_model_list: List[Type["PaitBaseResponseModel"]],
    target_pait_response_class: Optional[Type["PaitBaseResponseModel"]] = None,
    find_core_response_model: bool = False,
) -> Type["PaitBaseResponseModel"]:
    if target_pait_response_class:
        core_response_list: List[Type["PaitBaseResponseModel"]] = [
            i for i in response_model_list if i.is_core and issubclass(i, target_pait_response_class)
        ]
    else:
        core_response_list = [i for i in response_model_list if i.is_core]
    if find_core_response_model:
        if len(core_response_list) != 1:
            raise RuntimeError("Multiple pait response models were found")
        return core_response_list[0]
    else:
        return (core_response_list or response_model_list)[0]


def gen_example_value_from_python(obj: Any) -> Any:
    if isinstance(obj, dict):
        new_dict: dict = {}
        for key, value in obj.items():
            new_dict[key] = gen_example_value_from_python(value)
        return new_dict
    else:
        return python_type_default_value_dict.get(type(obj), obj)


def gen_example_dict_from_pydantic_base_model(
    pydantic_base_model: Type[BaseModel], use_example_value: bool = True
) -> dict:
    gen_dict: Dict[str, Any] = {}
    for key, model_field in pydantic_base_model.__fields__.items():
        if use_example_value:
            example_value: Any = model_field.field_info.extra.get("example", Undefined)
            if not isinstance(example_value, UndefinedType):
                gen_dict[key] = example_value_handle(example_value)
                continue

        if not isinstance(model_field.field_info.default, UndefinedType):
            gen_dict[key] = model_field.field_info.default
        elif model_field.field_info.default_factory and not isinstance(
            model_field.field_info.default_factory, UndefinedType
        ):
            gen_dict[key] = model_field.field_info.default_factory()
        else:
            annotation: Type = get_pydantic_annotation(key, pydantic_base_model)

            if inspect.isclass(annotation) and issubclass(annotation, BaseModel):
                gen_dict[key] = gen_example_dict_from_pydantic_base_model(annotation)
            else:
                sub_type: Optional[Type] = None
                try:
                    parse_typing_result: Union[List[Type[Any]], Type] = parse_typing(annotation)
                    if isinstance(parse_typing_result, list):
                        real_type: Type = parse_typing_result[0]
                    else:
                        real_type = parse_typing_result
                    if getattr(annotation, "__args__", None):
                        sub_type_set: Set[Type] = set(annotation.__args__)
                        if len(sub_type_set) == 1:
                            sub_type = sub_type_set.pop()
                except ParseTypeError:
                    real_type = annotation
                if (
                    real_type is list
                    and sub_type is not None
                    and inspect.isclass(sub_type)
                    and issubclass(sub_type, BaseModel)
                ):
                    gen_dict[key] = [gen_example_dict_from_pydantic_base_model(sub_type)]
                elif issubclass(real_type, Enum):
                    gen_dict[key] = [i for i in real_type.__members__.values()][0].value
                else:
                    gen_dict[key] = python_type_default_value_dict[real_type]
    return gen_dict


def gen_example_dict_from_schema(
    schema_dict: Dict[str, Any],
    definition_dict: Optional[dict] = None,
) -> Dict[str, Any]:
    gen_dict: Dict[str, Any] = {}
    if "properties" not in schema_dict:
        return gen_dict
    property_dict: Dict[str, Any] = schema_dict["properties"]
    if not definition_dict:
        _definition_dict: dict = schema_dict.get("definitions", {})
    else:
        _definition_dict = definition_dict
    for key, value in property_dict.items():
        if "items" in value and value["type"] == "array":
            if "$ref" in value["items"]:
                model_key: str = value["items"]["$ref"].split("/")[-1]
                gen_dict[key] = [gen_example_dict_from_schema(_definition_dict.get(model_key, {}), _definition_dict)]
            else:
                gen_dict[key] = []
        elif "$ref" in value:
            model_key = value["$ref"].split("/")[-1]
            gen_dict[key] = gen_example_dict_from_schema(_definition_dict.get(model_key, {}), _definition_dict)
        else:
            if "example" in value:
                gen_dict[key] = example_value_handle(value["example"])
            elif "default" in value:
                gen_dict[key] = value["default"]
            else:
                if "type" in value:
                    if value["type"] not in json_type_default_value_dict:
                        raise KeyError(f"Can not found type: {key} in json type")
                    gen_dict[key] = json_type_default_value_dict[value["type"]]
                else:
                    gen_dict[key] = "object"
            if isinstance(gen_dict[key], Enum):
                gen_dict[key] = gen_dict[key].value
    return gen_dict


def gen_example_json_from_schema(schema_dict: Dict[str, Any], cls: Optional[Type[JSONEncoder]] = None) -> str:
    return json.dumps(gen_example_dict_from_schema(schema_dict), cls=cls)


def get_parameter_list_from_pydantic_basemodel(pait_model: Type[BaseModel]) -> List["inspect.Parameter"]:
    """get class parameter list by attributes, if attributes not default value, it will be set `Undefined`"""
    parameter_list: Optional[List["inspect.Parameter"]] = getattr(pait_model, "_parameter_list", None)
    if parameter_list is not None:
        return parameter_list
    parameter_list = []
    for key, model_field in pait_model.__fields__.items():
        if not is_pait_field(model_field.field_info):
            raise TypeError(f"{model_field.field_info} must instance {BaseField} or {Depends}")  # pragma: no cover
        parameter = inspect.Parameter(
            key,
            inspect.Parameter.POSITIONAL_ONLY,
            default=model_field.field_info,
            annotation=get_pydantic_annotation(key, pait_model),
        )
        parameter_list.append(parameter)

    setattr(pait_model, "_parameter_list", parameter_list)
    return parameter_list


_class_parameter_list_dict: Dict[type, List["inspect.Parameter"]] = {}


def get_parameter_list_from_class(cbv_class: Type) -> List["inspect.Parameter"]:
    """get class parameter list by attributes, if attributes not default value, it will be set `Undefined`"""
    parameter_list: Optional[List["inspect.Parameter"]] = _class_parameter_list_dict.get(cbv_class)
    if parameter_list is not None:
        return parameter_list
    parameter_list = []
    if hasattr(cbv_class, "__annotations__"):
        for param_name, param_annotation in get_type_hints(cbv_class).items():
            default: Any = getattr(cbv_class, param_name, Undefined)
            if not is_pait_field(default):
                continue

            # Optimize parsing speed
            if not default.alias:
                default.request_key = param_name
            parameter: "inspect.Parameter" = inspect.Parameter(
                param_name,
                inspect.Parameter.POSITIONAL_ONLY,
                default=default,
                annotation=param_annotation,
            )
            parameter_list.append(parameter)
    _class_parameter_list_dict[cbv_class] = parameter_list
    return parameter_list


class CustomJSONEncoder(JSONEncoder):
    def default(self, obj: Any) -> Any:
        if isinstance(obj, datetime):
            return int(obj.timestamp())
        elif isinstance(obj, date):
            return obj.strftime("%Y-%m-%d")
        elif isinstance(obj, Decimal):
            return float(obj)
        elif isinstance(obj, Enum):
            return obj.value
        elif isinstance(obj, TemplateVar):
            obj = obj.get_value_from_template_context()
            if isinstance(obj, UndefinedType):
                obj = None
            return obj
        else:
            return super().default(obj)  # pragma: no cover



from typing import Any

from pait.plugin.check_json_resp import CheckJsonRespPlugin as _CheckJsonRespPlugin

from .base import JsonProtocol

__all__ = ["AsyncCheckJsonRespPlugin", "CheckJsonRespPlugin"]


class CheckJsonRespPlugin(JsonProtocol, _CheckJsonRespPlugin):
    def _sync_call(self, *args: Any, **kwargs: Any) -> Any:
        response: Any = super()._sync_call(*args, **kwargs)
        return self.gen_response(response)

    async def _async_call(self, *args: Any, **kwargs: Any) -> Any:
        response: Any = await super()._async_call(*args, **kwargs)
        return self.gen_response(response)


class AsyncCheckJsonRespPlugin(CheckJsonRespPlugin):
    """"""



import asyncio
import inspect
import logging
from contextlib import AbstractAsyncContextManager, AbstractContextManager
from types import TracebackType
from typing import TYPE_CHECKING, Any, Callable, Coroutine, Dict, Generator, List, Optional, Set, Tuple, Type, Union

from pydantic import BaseConfig, BaseModel
from pydantic.fields import Undefined, UndefinedType
from typing_extensions import Self  # type: ignore

from pait import field
from pait.app.base import BaseAppHelper
from pait.exceptions import (
    FieldValueTypeException,
    NotFoundFieldException,
    NotFoundValueException,
    PaitBaseException,
    ParseTypeError,
)
from pait.field import BaseField
from pait.g import pait_context
from pait.plugin.base import PluginProtocol
from pait.util import (
    FuncSig,
    create_pydantic_model,
    example_value_handle,
    gen_tip_exc,
    get_func_sig,
    get_parameter_list_from_class,
    get_parameter_list_from_pydantic_basemodel,
    ignore_pre_check,
    is_bounded_func,
)

if TYPE_CHECKING:
    from pait.model.core import ContextModel, PaitCoreModel


def raise_multiple_exc(exc_list: List[Exception]) -> None:
    """
    Multiple exceptions may be thrown during the parsing process, and these will be thrown one by one like a stack
    """
    if not exc_list:
        return
    try:
        raise exc_list.pop()
    finally:
        raise_multiple_exc(exc_list)


def parameter_2_dict(
    parameter_value_dict: Dict["inspect.Parameter", Any],
    pydantic_config: Type[BaseConfig],
    validators: Dict[str, classmethod] = None,
) -> Generator[dict, None, None]:
    """Convert all parameters into pydantic mods"""
    annotation_dict: Dict[str, Tuple[Type, Any]] = {}
    param_value_dict: Dict[str, Any] = {}

    # Resolve
    #   the key mismatch between Field.alias and request value
    #   different Fields but same alias
    key_set: Set[str] = set()

    for parameter, value in parameter_value_dict.items():
        param_field = parameter.default
        class_name = parameter.name
        value_key = parameter.default.request_key
        if value_key in key_set:
            yield create_pydantic_model(
                annotation_dict,
                pydantic_config=pydantic_config,
                pydantic_validators=validators,
            )(**param_value_dict).__dict__
            key_set = set()
            annotation_dict = {}
            param_value_dict = {}

        key_set.add(value_key)
        param_value_dict[value_key] = value
        annotation_dict[class_name] = (parameter.annotation, param_field)
    yield create_pydantic_model(
        annotation_dict,
        pydantic_config=pydantic_config,
        pydantic_validators=validators,
    )(**param_value_dict).__dict__


class BaseParamHandler(PluginProtocol):
    def __post_init__(self, pait_core_model: "PaitCoreModel", args: tuple, kwargs: dict) -> None:
        super(BaseParamHandler, self).__post_init__(pait_core_model, args, kwargs)

        # cbv handle
        context_model: ContextModel = pait_context.get()
        self.cbv_instance: Optional[Any] = context_model.cbv_instance
        self._app_helper: BaseAppHelper = context_model.app_helper
        self.cbv_type: Optional[Type] = None

        if self.cbv_instance:
            self.cbv_type = self.cbv_instance.__class__
            self.cbv_param_list: List["inspect.Parameter"] = get_parameter_list_from_class(self.cbv_type)
        else:
            self.cbv_param_list = []
            # cbv_type = getattr(inspect.getmodule(func), func.__qualname__.split(".<locals>", 1)[0].rsplit(".", 1)[0])

        self.pre_depend_list: List[Callable] = pait_core_model.pre_depend_list
        self.pydantic_model_config: Type[BaseConfig] = pait_core_model.pydantic_model_config

    @classmethod
    def check_depend_handle(cls, func: Callable) -> Any:
        if inspect.ismethod(func) and not is_bounded_func(func):
            raise ValueError(f"Method: {func.__qualname__} is not a bounded function")  # pragma: no cover
        func_sig: FuncSig = get_func_sig(func)  # get and cache func sig
        cls.check_param_field_handle(func_sig, func_sig.param_list)

    @staticmethod
    def check_field_type(value: Any, target_type: Any, error_msg: str) -> None:
        if isinstance(value, UndefinedType):
            return
        if getattr(value, "__call__", None):
            value = value()
        try:
            create_pydantic_model({"faker_param_name": (target_type, ...)})(faker_param_name=value)
        except Exception:
            raise ParseTypeError(error_msg)

    @classmethod
    def check_param_field_by_parameter(
        cls,
        parameter: inspect.Parameter,
    ) -> None:
        if isinstance(parameter.default, field.Depends):
            cls.check_depend_handle(parameter.default.func)
        elif isinstance(parameter.default, field.BaseField):
            if not parameter.default.alias:
                parameter.default.request_key = parameter.name
            if not ignore_pre_check:
                if parameter.default.alias and not isinstance(parameter.default.alias, str):
                    raise FieldValueTypeException(
                        parameter.name,
                        f"{parameter.name}'s Field.alias type must str. value: {parameter.default.alias}",
                    )
                try:
                    cls.check_field_type(
                        parameter.default.default,
                        parameter.annotation,
                        f"{parameter.name}'s Field.default type must {parameter.annotation}."
                        f" value:{parameter.default.default}",
                    )
                    if parameter.default.default_factory:
                        cls.check_field_type(
                            parameter.default.default_factory(),
                            parameter.annotation,
                            f"{parameter.name}'s Field.default_factory type must {parameter.annotation}."
                            f" value:{parameter.default.default_factory()}",
                        )
                    example_value: Any = parameter.default.extra.get("example", Undefined)
                    cls.check_field_type(
                        example_value_handle(example_value),
                        parameter.annotation,
                        f"{parameter.name}'s Field.example type must {parameter.annotation} not {example_value}",
                    )
                except ParseTypeError as e:
                    raise FieldValueTypeException(parameter.name, str(e))
        else:
            raise NotFoundFieldException(parameter.name, f"{parameter.name}'s Field not found")  # pragma: no cover

    @classmethod
    def check_param_field_handle(
        cls,
        _object: Union[FuncSig, Type, None],
        param_list: List["inspect.Parameter"],
    ) -> None:

        for parameter in param_list:
            try:
                if parameter.default != parameter.empty:
                    # kwargs param
                    # support model: def demo(pydantic.BaseModel: BaseModel = pait.field.BaseField())
                    cls.check_param_field_by_parameter(parameter)
                else:
                    # args param
                    # support model: model: ModelType
                    if parameter.annotation == Self:
                        continue
                    if not issubclass(parameter.annotation, BaseModel):
                        continue
                    # cache and get parameter_list
                    param_list = get_parameter_list_from_pydantic_basemodel(parameter.annotation)
                    for _parameter in param_list:
                        cls.check_param_field_by_parameter(_parameter)
            except PaitBaseException as e:
                raise gen_tip_exc(_object, e, parameter)

    @classmethod
    def pre_hook(cls, pait_core_model: "PaitCoreModel", kwargs: Dict) -> None:
        # check and load param from pre depend
        for pre_depend in pait_core_model.pre_depend_list:
            cls.check_depend_handle(pre_depend)

        # check and load param from func
        func_sig: FuncSig = get_func_sig(pait_core_model.func)
        cls.check_param_field_handle(func_sig, func_sig.param_list)

        # TODO support cbv class Attribute
        # I don't know how to get the class of the decorated function at the initialization of the decorator,
        # which may be an unattainable requirement

    @classmethod
    def pre_load_hook(cls, pait_core_model: "PaitCoreModel", kwargs: Dict) -> Dict:
        super().pre_load_hook(pait_core_model, kwargs)
        if ignore_pre_check:
            # pre_check has helped to do the same task as pre_load
            cls.pre_hook(pait_core_model, kwargs)  # pragma: no cover
        return kwargs

    @classmethod
    def pre_check_hook(cls, pait_core_model: "PaitCoreModel", kwargs: Dict) -> None:
        super().pre_check_hook(pait_core_model, kwargs)
        cls.pre_hook(pait_core_model, kwargs)

    def _set_parameter_value_to_args(self, parameter: inspect.Parameter, func_args: list) -> bool:
        """Extract the self parameter of the cbv handler,
        the request parameter of the route and the parameter of type PaitBaseModel,
        and check if there are any other parameters that do not meet the conditions

        Sort by frequency of occurrence
        """
        if self.cbv_instance and (
            parameter.annotation == Self or (not func_args and parameter.annotation == parameter.empty)
        ):
            # first parma must self param, looking forward to the appearance of `self type
            func_args.append(self._app_helper.cbv_instance)
        elif issubclass(parameter.annotation, BaseModel):
            return True
        elif self._app_helper.check_request_type(parameter.annotation):
            # support request param(def handle(request: Request))
            func_args.append(self._app_helper.request)
        else:
            logging.warning(f"Pait not support args: {parameter}")  # pragma: no cover
        return False

    @staticmethod
    def request_value_handle(
        parameter: inspect.Parameter,
        request_value: Any,
        base_model_dict: Dict[str, Any],
        parameter_value_dict: Dict["inspect.Parameter", Any],
    ) -> None:
        """parse request_value and set to base_model_dict or parameter_value_dict"""
        if request_value is None:
            raise NotFoundValueException(parameter.name, f"Can not found {parameter.name} value")
        pait_field: BaseField = parameter.default
        annotation: Type[BaseModel] = parameter.annotation

        # some type like dict, but not isinstance Mapping, e.g: werkzeug.datastructures.EnvironHeaders
        # assert getattr(request_value, "get", None), f"{parameter.name}'s request value must like dict"
        if not pait_field.raw_return:

            request_value = pait_field.request_value_handle(request_value)
            if request_value is Undefined:
                raise NotFoundValueException(parameter.name, f"Can not found {parameter.name} value")

        if inspect.isclass(annotation) and issubclass(annotation, BaseModel):
            # parse annotation is pydantic.BaseModel and base_model_dict not None
            base_model_dict[parameter.name] = annotation(**request_value)
        else:
            # parse annotation is python type and pydantic.field
            parameter_value_dict[parameter] = request_value

    def get_request_value_from_parameter(self, parameter: inspect.Parameter) -> Union[Any, Coroutine]:
        field_name: str = parameter.default.get_field_name()
        # Note: not use hasattr with LazyProperty (
        #   because hasattr will calling getattr(obj, name) and catching AttributeError,
        # )
        app_field_func: Optional[Callable] = getattr(self._app_helper, field_name, None)
        if app_field_func is None:
            raise NotFoundFieldException(
                parameter.name, f"field: {field_name} not found in {self._app_helper}"
            )  # pragma: no cover
        return app_field_func()

    def valid_and_merge_kwargs_by_single_field_dict(
        self,
        single_field_dict: Dict["inspect.Parameter", Any],
        kwargs_param_dict: Dict[str, Any],
        _object: Union[FuncSig, Type, None],
    ) -> None:
        try:
            for parse_dict in parameter_2_dict(single_field_dict, self.pydantic_model_config):
                kwargs_param_dict.update(parse_dict)
        except Exception as e:
            raise e from gen_tip_exc(_object, e)

    @staticmethod
    def valid_and_merge_kwargs_by_pydantic_model(
        single_field_dict: Dict["inspect.Parameter", Any],
        kwargs_param_dict: Dict[str, Any],
        pydantic_model: Type[BaseModel],
        _object: Union[FuncSig, Type, None],
    ) -> None:
        try:
            request_dict: Dict[str, Any] = kwargs_param_dict.copy()
            for k, v in single_field_dict.items():
                request_dict[k.default.request_key] = v
            kwargs_param_dict.update(pydantic_model(**request_dict))
        except Exception as e:
            raise e from gen_tip_exc(_object, e)


class ParamHandler(BaseParamHandler):
    def __init__(self, **kwargs: Any) -> None:
        super().__init__(**kwargs)
        self._contextmanager_list: List[AbstractContextManager] = []

    def param_handle(
        self,
        _object: Union[FuncSig, Type, None],
        param_list: List["inspect.Parameter"],
        pydantic_model: Type[BaseModel] = None,
    ) -> Tuple[List[Any], Dict[str, Any]]:
        args_param_list: List[Any] = []
        kwargs_param_dict: Dict[str, Any] = {}

        single_field_dict: Dict["inspect.Parameter", Any] = {}

        for parameter in param_list:
            try:
                if parameter.default != parameter.empty:
                    # kwargs param
                    # support model: def demo(pydantic.BaseModel: BaseModel = pait.field.BaseField())
                    if isinstance(parameter.default, field.Depends):
                        kwargs_param_dict[parameter.name] = self._depend_handle(parameter.default.func)
                    else:
                        request_value: Any = self.get_request_value_from_parameter(parameter)
                        self.request_value_handle(parameter, request_value, kwargs_param_dict, single_field_dict)
                else:
                    # args param
                    # support model: model: ModelType
                    self.set_parameter_value_to_args(parameter, args_param_list)
            except PaitBaseException as e:
                raise gen_tip_exc(_object, e, parameter)
        # support field: def demo(demo_param: int = pait.field.BaseField())
        if single_field_dict:
            if pydantic_model:
                self.valid_and_merge_kwargs_by_pydantic_model(
                    single_field_dict, kwargs_param_dict, pydantic_model, _object
                )
            else:
                self.valid_and_merge_kwargs_by_single_field_dict(single_field_dict, kwargs_param_dict, _object)
        return args_param_list, kwargs_param_dict

    def set_parameter_value_to_args(self, parameter: inspect.Parameter, func_args: list) -> None:
        """use func_args param faster return and extend func_args"""
        if not self._set_parameter_value_to_args(parameter, func_args):
            return
        # support pait_model param(def handle(model: PaitBaseModel))
        _pait_model: Type[BaseModel] = parameter.annotation
        _, kwargs = self.param_handle(
            None, get_parameter_list_from_pydantic_basemodel(_pait_model), pydantic_model=_pait_model
        )
        # Data has been validated or is from a trusted source
        func_args.append(_pait_model.construct(**kwargs))

    def _depend_handle(self, func: Callable) -> Any:
        func_sig: FuncSig = get_func_sig(func)
        _func_args, _func_kwargs = self.param_handle(func_sig, func_sig.param_list)
        func_result: Any = func(*_func_args, **_func_kwargs)
        if isinstance(func_result, AbstractContextManager):
            self._contextmanager_list.append(func_result)
            return func_result.__enter__()
        else:
            return func_result

    def _gen_param(self) -> None:
        # check param from pre depend
        for pre_depend in self.pre_depend_list:
            self._depend_handle(pre_depend)

        # gen and check param from func
        func_sig: FuncSig = get_func_sig(self.pait_core_model.func)
        self.args, self.kwargs = self.param_handle(func_sig, func_sig.param_list)

        # gen and check param from class
        if self.cbv_param_list and self.cbv_type:
            _, kwargs = self.param_handle(self.cbv_type, self.cbv_param_list)
            self.cbv_instance.__dict__.update(kwargs)
        return None

    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        try:
            with self:
                return self.call_next(*self.args, **self.kwargs)
        except Exception as e:
            raise e from gen_tip_exc(self.call_next, e)

    def __enter__(self) -> "ParamHandler":
        self._gen_param()
        return self

    def __exit__(
        self, exc_type: Optional[Type[BaseException]], exc_val: Optional[BaseException], exc_tb: Optional[TracebackType]
    ) -> Optional[bool]:
        exc_list: List[Exception] = []
        for contextmanager in self._contextmanager_list:
            try:
                contextmanager.__exit__(exc_type, exc_val, exc_tb)
            except Exception as e:
                exc_list.append(e)
            else:
                if exc_type and issubclass(exc_type, Exception):
                    exc_list.append(exc_type(exc_val).with_traceback(exc_tb))
        if exc_list:
            raise_multiple_exc(exc_list)
            return True
        else:
            return False


class AsyncParamHandler(BaseParamHandler):
    def __init__(self, **kwargs: Any) -> None:
        super().__init__(**kwargs)
        self._contextmanager_list: List[Union[AbstractAsyncContextManager, AbstractContextManager]] = []

    async def param_handle(
        self,
        _object: Union[FuncSig, Type],
        param_list: List["inspect.Parameter"],
        pydantic_model: Type[BaseModel] = None,
    ) -> Tuple[List[Any], Dict[str, Any]]:
        args_param_list: List[Any] = []
        kwargs_param_dict: Dict[str, Any] = {}
        single_field_dict: Dict["inspect.Parameter", Any] = {}

        async def _param_handle(parameter: inspect.Parameter) -> None:
            try:
                if parameter.default != parameter.empty:
                    # kwargs param
                    # support like: def demo(pydantic.BaseModel: BaseModel = pait.field.BaseField())
                    if isinstance(parameter.default, field.Depends):
                        kwargs_param_dict[parameter.name] = await self._depend_handle(parameter.default.func)
                    else:
                        request_value: Any = self.get_request_value_from_parameter(parameter)
                        if asyncio.iscoroutine(request_value) or asyncio.isfuture(request_value):
                            request_value = await request_value
                        self.request_value_handle(parameter, request_value, kwargs_param_dict, single_field_dict)
                else:
                    # args param
                    # support model: model: ModelType
                    await self.set_parameter_value_to_args(_object, parameter, args_param_list)
            except PaitBaseException as closer_e:
                raise gen_tip_exc(_object, closer_e, parameter)

        if param_list:
            await asyncio.gather(*[_param_handle(parameter) for parameter in param_list])
        # support field: def demo(demo_param: int = pait.field.BaseField())
        if single_field_dict:
            if single_field_dict:
                if pydantic_model:
                    self.valid_and_merge_kwargs_by_pydantic_model(
                        single_field_dict, kwargs_param_dict, pydantic_model, _object
                    )
                else:
                    self.valid_and_merge_kwargs_by_single_field_dict(single_field_dict, kwargs_param_dict, _object)
        return args_param_list, kwargs_param_dict

    async def set_parameter_value_to_args(
        self, _object: Union[FuncSig, Type], parameter: inspect.Parameter, func_args: list
    ) -> None:
        """use func_args param faster return and extend func_args"""
        if not self._set_parameter_value_to_args(parameter, func_args):
            return
        _pait_model: Type[BaseModel] = parameter.annotation
        # Data has been validated or is from a trusted source
        _, kwargs = await self.param_handle(
            _object, get_parameter_list_from_pydantic_basemodel(_pait_model), _pait_model
        )
        func_args.append(_pait_model.construct(**kwargs))

    async def _depend_handle(self, func: Callable) -> Any:
        func_sig: FuncSig = get_func_sig(func)
        _func_args, _func_kwargs = await self.param_handle(func_sig, func_sig.param_list)
        func_result: Any = func(*_func_args, **_func_kwargs)
        if asyncio.iscoroutine(func_result):
            func_result = await func_result
        if isinstance(func_result, AbstractAsyncContextManager):
            self._contextmanager_list.append(func_result)
            return await func_result.__aenter__()
        elif isinstance(func_result, AbstractContextManager):
            self._contextmanager_list.append(func_result)
            return func_result.__enter__()
        else:
            return func_result

    async def _gen_param(self) -> None:
        # check param from pre depend
        if self.pre_depend_list:
            await asyncio.gather(*[self._depend_handle(pre_depend) for pre_depend in self.pre_depend_list])

        # gen and check param from func
        func_sig: FuncSig = get_func_sig(self.pait_core_model.func)
        self.args, self.kwargs = await self.param_handle(func_sig, func_sig.param_list)

        # gen and check param from class
        if self.cbv_param_list and self.cbv_type:
            _, kwargs = await self.param_handle(self.cbv_type, self.cbv_param_list)
            self.cbv_instance.__dict__.update(kwargs)
        return None

    async def __call__(self, *args: Any, **kwargs: Any) -> Any:
        async with self:
            return await self.call_next(*self.args, **self.kwargs)

    async def __aenter__(self) -> "AsyncParamHandler":
        try:
            await self._gen_param()
            return self
        except Exception as e:
            raise e from gen_tip_exc(self.call_next, e)

    async def __aexit__(
        self, exc_type: Optional[Type[BaseException]], exc_val: Optional[BaseException], exc_tb: Optional[TracebackType]
    ) -> Optional[bool]:
        exc_list: List[Exception] = []

        async def _aexit(contextmanager: Union[AbstractAsyncContextManager, AbstractContextManager]) -> None:
            try:
                if isinstance(contextmanager, AbstractContextManager):
                    contextmanager.__exit__(exc_type, exc_val, exc_tb)
                else:
                    await contextmanager.__aexit__(exc_type, exc_val, exc_tb)
            except Exception as e:
                exc_list.append(e)
            else:
                if exc_type and issubclass(exc_type, Exception):
                    exc_list.append(exc_type(exc_val).with_traceback(exc_tb))

        if self._contextmanager_list:
            await asyncio.gather(*[_aexit(contextmanager) for contextmanager in self._contextmanager_list])
        if exc_list:
            raise_multiple_exc(exc_list)
            return True
        else:
            return False



import functools
import inspect
import logging
from typing import Callable, Dict, Optional, Type, Union

from starlette import routing
from starlette.applications import Starlette
from starlette.endpoints import HTTPEndpoint

from pait.g import pait_data
from pait.model.core import PaitCoreModel
from pait.util import http_method_tuple

from ._app_helper import AppHelper

__all__ = ["load_app"]


def _load_route(
    *,
    route: routing.Route,
    _pait_data: Dict[str, PaitCoreModel],
    project_name: str,
    prefix_path: Optional[str] = None,
    auto_load_route: bool = False,
) -> None:
    path: str = route.path
    if prefix_path:
        path = prefix_path + path
    openapi_path: str = path
    method_set: set = route.methods or set()
    route_name: str = route.name
    endpoint: Union[Callable, Type] = route.endpoint
    pait_id: str = getattr(route.endpoint, "_pait_id", "")
    if not pait_id and inspect.isclass(endpoint) and issubclass(endpoint, HTTPEndpoint):  # type: ignore
        for method in http_method_tuple:
            method_endpoint = getattr(endpoint, method, None)
            if not method_endpoint:
                continue
            method_set = {method}
            pait_id = getattr(method_endpoint, "_pait_id", None)
            if not pait_id:
                if not auto_load_route:
                    logging.warning(f"{route_name}.{method} can not found pait id")  # pragma: no cover
                from pait.app.starlette import pait

                method_endpoint = pait()(method_endpoint)
                pait_id = getattr(method_endpoint, "_pait_id", None)
                setattr(endpoint, method, method_endpoint)

            pait_data.add_route_info(
                AppHelper.app_name, pait_id, path, openapi_path, method_set, f"{route_name}.{method}", project_name
            )
            _pait_data[pait_id] = pait_data.get_pait_data(AppHelper.app_name, pait_id)
    elif pait_id:
        pait_data.add_route_info(AppHelper.app_name, pait_id, path, openapi_path, method_set, route_name, project_name)
        _pait_data[pait_id] = pait_data.get_pait_data(AppHelper.app_name, pait_id)
    elif auto_load_route:

        from pait.app.starlette import pait

        endpoint = pait()(endpoint)
        pait_id = getattr(endpoint, "_pait_id", "")

        route.endpoint = endpoint
        endpoint_handler = endpoint
        while isinstance(endpoint_handler, functools.partial):
            endpoint_handler = endpoint_handler.func
        if inspect.isfunction(endpoint_handler) or inspect.ismethod(endpoint_handler):
            # Endpoint is function or method. Treat it as `func(request) -> response`.
            route.app = routing.request_response(endpoint)
        else:
            # Endpoint is a class. Treat it as ASGI.
            route.app = endpoint

        pait_data.add_route_info(AppHelper.app_name, pait_id, path, openapi_path, method_set, route_name, project_name)
        _pait_data[pait_id] = pait_data.get_pait_data(AppHelper.app_name, pait_id)
    else:
        logging.warning(f"{route_name} can not found pait id")  # pragma: no cover


def load_app(app: Starlette, project_name: str = "", auto_load_route: bool = False) -> Dict[str, PaitCoreModel]:
    """Read data from the route that has been registered to `pait`"""
    _pait_data: Dict[str, PaitCoreModel] = {}
    for route in app.routes:
        if isinstance(route, routing.Route):
            _load_route(route=route, _pait_data=_pait_data, project_name=project_name, auto_load_route=auto_load_route)
        elif isinstance(route, routing.Mount):
            for sub_route in route.routes:
                _load_route(
                    route=sub_route,  # type: ignore
                    _pait_data=_pait_data,
                    project_name=project_name,
                    prefix_path=route.path,
                    auto_load_route=auto_load_route,
                )
        else:
            logging.info(f"load_app func not support route:{route.__class__}")

    return _pait_data



from typing import Any, Callable, Dict, Optional

from sanic import HTTPResponse
from sanic.response import json

from pait.plugin.base import PluginManager, PluginProtocol


class JsonProtocol(PluginProtocol):
    status: int
    headers: Optional[Dict[str, str]]
    content_type: str
    dumps: Optional[Callable[..., str]]
    json_kwargs: dict

    def gen_response(self, response: Any) -> HTTPResponse:
        return json(
            response,
            status=self.status,
            headers=self.headers,
            content_type=self.content_type,
            dumps=self.dumps,
            **self.json_kwargs if self.json_kwargs else {},
        )

    @classmethod
    def build(
        cls,
        status: Optional[int] = None,
        headers: Optional[Dict[str, str]] = None,
        content_type: Optional[str] = None,
        dumps: Optional[Callable[..., str]] = None,
        json_kwargs: Optional[dict] = None,
        **kwargs: Any,
    ) -> "PluginManager":
        return super().build(
            status=status or 200,
            headers=headers or {},
            content_type=content_type or "application/json",
            dumps=dumps,
            json_kwargs=json_kwargs,
            **kwargs,
        )



from typing import Any, Dict, Tuple

from tornado.web import RequestHandler

from pait.plugin.cache_response import CacheResponsePlugin as _CacheResponsePlugin

__all__ = ["CacheResponsePlugin"]


class CacheResponsePlugin(_CacheResponsePlugin):
    def _gen_key(self, *args: Any, **kwargs: Any) -> Tuple[str, str]:
        return super()._gen_key(*args[1:], **kwargs)

    def _dumps(self, response: Any, *args: Any, **kwargs: Any) -> Any:
        tornado_handle: RequestHandler = args[0]
        if isinstance(response, Exception):
            return super()._dumps(response, *args, **kwargs)
        cache_dict: Dict[str, Any] = {
            "write_buffer": tornado_handle._write_buffer,
            "status_code": tornado_handle._status_code,
            "headers": tornado_handle._headers,
        }
        return super()._dumps(cache_dict, *args, **kwargs)

    def _loads(self, response: Any, *args: Any, **kwargs: Any) -> Any:
        response = super()._loads(response, *args, **kwargs)
        if isinstance(response, Exception):
            return response
        tornado_handle: RequestHandler = args[0]
        tornado_handle._write_buffer = response["write_buffer"]
        tornado_handle._status_code = response["status_code"]
        tornado_handle._headers = response["headers"]
        return tornado_handle



import json
from typing import Any, Dict, List, Optional
from urllib.parse import urlencode

from sanic import response
from sanic.app import Sanic
from sanic.blueprints import Blueprint
from sanic.exceptions import NotFound
from sanic.request import Request
from sanic.response import HTTPResponse
from sanic_testing.testing import SanicTestClient, TestingResponse  # type: ignore

from pait.api_doc.html import get_redoc_html as _get_redoc_html
from pait.api_doc.html import get_swagger_ui_html as _get_swagger_ui_html
from pait.api_doc.open_api import PaitOpenAPI
from pait.app.base.doc_route import AddDocRoute as _AddDocRoute
from pait.app.base.doc_route import DocHtmlRespModel, OpenAPIRespModel
from pait.field import Depends
from pait.g import config
from pait.model.core import PaitCoreModel
from pait.model.template import TemplateContext

from ._load_app import load_app
from ._pait import Pait

__all__ = ["add_doc_route", "AddDocRoute"]


class AddDocRoute(_AddDocRoute[Sanic]):
    not_found_exc: Exception = NotFound("")

    def _gen_route(self, app: Sanic) -> Any:
        doc_pait = self._get_doc_pait(Pait)

        def _get_open_json_url(request: Request, r_pin_code: str, url_dict: Dict[str, Any]) -> str:
            _scheme: str = self.scheme or request.scheme
            if self.open_json_url_only_path:
                openapi_json_url: str = f"{'/'.join(request.path.split('/')[:-1])}/openapi.json"
            else:
                openapi_json_url = f"{_scheme}://{request.host}{'/'.join(request.path.split('/')[:-1])}/openapi.json"
            if r_pin_code:
                url_dict["pin_code"] = r_pin_code
            openapi_json_url += "?" + urlencode(url_dict)
            return openapi_json_url

        @doc_pait(response_model_list=[DocHtmlRespModel])
        def get_redoc_html(
            request: Request,
            r_pin_code: str = Depends.i(self._get_request_pin_code),
            url_dict: Dict[str, Any] = Depends.i(self._get_request_template_map()),
        ) -> HTTPResponse:
            return response.html(
                _get_redoc_html(
                    _get_open_json_url(request, r_pin_code, url_dict),
                    src_url=self.redoc_src_url,
                    title=self.title,
                )
            )

        @doc_pait(response_model_list=[DocHtmlRespModel])
        def get_swagger_ui_html(
            request: Request,
            r_pin_code: str = Depends.i(self._get_request_pin_code),
            url_dict: Dict[str, Any] = Depends.i(self._get_request_template_map()),
        ) -> HTTPResponse:
            return response.html(
                _get_swagger_ui_html(
                    _get_open_json_url(request, r_pin_code, url_dict),
                    title=self.title,
                    swagger_ui_bundle=self.swagger_ui_bundle,
                    swagger_ui_standalone_preset=self.swagger_ui_standalone_preset,
                    swagger_ui_url=self.swagger_ui_url,
                )
            )

        @doc_pait(pre_depend_list=[self._get_request_pin_code], response_model_list=[OpenAPIRespModel])
        def openapi_route(
            request: Request, url_dict: Dict[str, Any] = Depends.i(self._get_request_template_map(extra_key=True))
        ) -> HTTPResponse:
            pait_dict: Dict[str, PaitCoreModel] = load_app(request.app, project_name=self.project_name)
            _scheme: str = self.scheme or request.scheme
            with TemplateContext(url_dict):
                pait_openapi: PaitOpenAPI = PaitOpenAPI(
                    pait_dict,
                    title=self.title,
                    open_api_server_list=[{"url": f"{_scheme}://{request.host}", "description": ""}],
                    open_api_tag_list=self.open_api_tag_list,
                )
                return response.json(pait_openapi.open_api_dict, dumps=json.dumps, cls=config.json_encoder)

        blueprint: Blueprint = Blueprint(self.title, self.prefix)
        blueprint.add_route(get_redoc_html, "/redoc", methods={"GET"})
        blueprint.add_route(get_swagger_ui_html, "/swagger", methods={"GET"})
        blueprint.add_route(openapi_route, "/openapi.json", methods={"GET"})
        app.blueprint(blueprint)


def add_doc_route(
    app: Sanic,
    scheme: Optional[str] = None,
    open_json_url_only_path: bool = False,
    prefix: str = "",
    pin_code: str = "",
    title: str = "",
    open_api_tag_list: Optional[List[Dict[str, Any]]] = None,
    project_name: str = "",
) -> None:
    AddDocRoute(
        scheme=scheme,
        open_json_url_only_path=open_json_url_only_path,
        prefix=prefix,
        pin_code=pin_code,
        title=title,
        open_api_tag_list=open_api_tag_list,
        project_name=project_name,
        app=app,
    )



import logging
from typing import Any, Callable, Dict, Generic, List, Optional, Type, TypeVar

from pydantic import BaseModel

from pait.core import Pait
from pait.field import Query
from pait.g import config
from pait.model.response import PaitHtmlResponseModel, PaitJsonResponseModel
from pait.model.status import PaitStatus
from pait.model.tag import Tag

logger: logging.Logger = logging.getLogger(__name__)
__all__ = ["DocHtmlRespModel", "OpenAPIRespModel", "AddDocRoute"]


class DocHtmlRespModel(PaitHtmlResponseModel):
    header: dict = {"X-Example-Type": "html"}
    description: str = "doc html response"


class OpenAPIRespModel(PaitJsonResponseModel):
    class OpenAPIResponseModel(BaseModel):
        pass

    description: str = "open api json response"
    # TODO replace to openapi response model
    response_data: dict = {}  # type: ignore


APP_T = TypeVar("APP_T")


class AddDocRoute(Generic[APP_T]):
    not_found_exc: Exception

    def __init__(
        self,
        scheme: Optional[str] = None,
        open_json_url_only_path: bool = False,
        prefix: str = "",
        pin_code: str = "",
        title: str = "",
        open_api_tag_list: Optional[List[Dict[str, Any]]] = None,
        project_name: str = "",
        app: APP_T = None,
        src_url: Optional[str] = None,
        swagger_ui_url: Optional[str] = None,
        swagger_ui_bundle: Optional[str] = None,
        swagger_ui_standalone_preset: Optional[str] = None,
    ):
        if pin_code:
            logging.info(f"doc route start pin code:{pin_code}")
        self.scheme: Optional[str] = scheme
        self.open_json_url_only_path: bool = open_json_url_only_path
        self.prefix: str = prefix or "/"
        self.pin_code: str = pin_code
        self.title: str = title or "Pait Doc"
        self.open_api_tag_list: Optional[List[Dict[str, Any]]] = open_api_tag_list
        self.project_name: str = project_name
        self._is_gen: bool = False
        self.redoc_src_url: Optional[str] = None
        self.swagger_ui_url: Optional[str] = None
        self.swagger_ui_bundle: Optional[str] = None
        self.swagger_ui_standalone_preset: Optional[str] = None

        if app:
            self._is_gen = True
            self._gen_route(app)

    def _get_request_pin_code(self, r_pin_code: str = Query.i("", alias="pin_code")) -> Optional[str]:
        if self.pin_code:
            if r_pin_code != self.pin_code:
                raise self.not_found_exc
        return r_pin_code

    @staticmethod
    def _get_request_template_map(extra_key: bool = False) -> Callable:
        def _get_request_template_map(
            url_dict: Dict[str, Any] = Query.i(
                default_factory=dict,
                raw_return=True,
                description="Set the template variable, for example, there is a template parameter token, "
                "then the requested parameter is template-token=xxx",
            )
        ) -> dict:
            template_dict: Dict[str, Any] = {}
            for k, v in url_dict.items():
                if k.startswith("template-"):
                    if extra_key:
                        k = k.replace("template-", "")
                    template_dict[k] = v
            return template_dict

        return _get_request_template_map

    @staticmethod
    def _get_doc_pait(pait_class: Type[Pait]) -> Pait:
        return pait_class(
            author=config.author or ("so1n",),
            status=config.status or PaitStatus.release,
            tag=(Tag("pait_doc", desc="pait default doc route"),),
            group="pait_doc",
        )

    def _gen_route(self, app: APP_T) -> Any:  # type: ignore
        raise NotImplementedError()

    def gen_route(self, app: APP_T) -> None:
        """Will remove on version 1.0"""
        if self._is_gen:
            raise RuntimeError("Doc route has been generated")
        self._gen_route(app)
        self._is_gen = True



from abc import ABC
from typing import Any, Dict, List, Optional
from urllib.parse import urlencode

from tornado.web import Application, RequestHandler

from pait.api_doc.html import get_redoc_html as _get_redoc_html
from pait.api_doc.html import get_swagger_ui_html as _get_swagger_ui_html
from pait.api_doc.open_api import PaitOpenAPI
from pait.app.base.doc_route import AddDocRoute as _AddDocRoute
from pait.field import Depends
from pait.model.core import PaitCoreModel
from pait.model.template import TemplateContext

from ._load_app import load_app
from ._pait import Pait

__all__ = ["add_doc_route", "AddDocRoute"]


class NotFound(Exception):
    pass


class AddDocRoute(_AddDocRoute[Application]):
    not_found_exc: Exception = NotFound()

    def _gen_route(self, app: Application) -> Any:
        doc_pait = self._get_doc_pait(Pait)
        doc_route_self: "AddDocRoute" = self

        class BaseHandle(RequestHandler, ABC):
            def _handle_request_exception(self, exc: BaseException) -> None:
                if isinstance(exc, NotFound):
                    self.set_status(404)
                    self.write(
                        (
                            "The requested URL was not found on the server. If you entered"
                            " the URL manually please check your spelling and try again."
                        )
                    )
                    self.finish()
                else:
                    self.set_status(500)  # pragma: no cover
                    self.finish()  # pragma: no cover

            def _get_open_json_url(self, r_pin_code: str, url_dict: Dict[str, Any]) -> str:
                _scheme: str = doc_route_self.scheme or self.request.protocol
                if doc_route_self.open_json_url_only_path:
                    openapi_json_url: str = f"{'/'.join(self.request.path.split('/')[:-1])}/openapi.json"
                else:
                    openapi_json_url = (
                        f"{_scheme}://{self.request.host}{'/'.join(self.request.path.split('/')[:-1])}/openapi.json"
                    )
                if r_pin_code:
                    url_dict["pin_code"] = r_pin_code
                openapi_json_url += "?" + urlencode(url_dict)
                return openapi_json_url

        class GetRedocHtmlHandle(BaseHandle, ABC):
            @doc_pait()
            async def get(
                self,
                r_pin_code: str = Depends.i(doc_route_self._get_request_pin_code),
                url_dict: Dict[str, Any] = Depends.i(self._get_request_template_map()),
            ) -> None:
                self.write(
                    _get_redoc_html(
                        self._get_open_json_url(r_pin_code, url_dict),
                        src_url=doc_route_self.redoc_src_url,
                        title=doc_route_self.title,
                    )
                )

        class GetSwaggerUiHtmlHandle(BaseHandle, ABC):
            @doc_pait()
            async def get(
                self,
                r_pin_code: str = Depends.i(doc_route_self._get_request_pin_code),
                url_dict: Dict[str, Any] = Depends.i(self._get_request_template_map()),
            ) -> None:
                self.write(
                    _get_swagger_ui_html(
                        self._get_open_json_url(r_pin_code, url_dict),
                        title=doc_route_self.title,
                        swagger_ui_bundle=doc_route_self.swagger_ui_bundle,
                        swagger_ui_standalone_preset=doc_route_self.swagger_ui_standalone_preset,
                        swagger_ui_url=doc_route_self.swagger_ui_url,
                    )
                )

        class OpenApiHandle(BaseHandle, ABC):
            @doc_pait(pre_depend_list=[doc_route_self._get_request_pin_code])
            async def get(
                self, url_dict: Dict[str, Any] = Depends.i(self._get_request_template_map(extra_key=True))
            ) -> None:
                pait_dict: Dict[str, PaitCoreModel] = load_app(
                    self.application, project_name=doc_route_self.project_name
                )
                _scheme: str = doc_route_self.scheme or self.request.protocol
                with TemplateContext(url_dict):
                    pait_openapi: PaitOpenAPI = PaitOpenAPI(
                        pait_dict,
                        title=doc_route_self.title,
                        open_api_server_list=[{"url": f"{_scheme}://{self.request.host}", "description": ""}],
                        open_api_tag_list=doc_route_self.open_api_tag_list,
                    )
                    self.set_header("Content-Type", "application/json; charset=UTF-8")
                    self.write(pait_openapi.content)

        prefix: str = doc_route_self.prefix
        if not prefix.endswith("/"):
            prefix = prefix + "/"
        # Method 1
        # app.add_handlers(
        #     r".*$",
        #     [
        #         (r"{}redoc".format(prefix), GetRedocHtmlHandle),
        #         (r"{}swagger".format(prefix), GetSwaggerUiHtmlHandle),
        #         (r"{}openapi.json".format(prefix), OpenApiHandle),
        #     ]
        # )
        #
        # Method 2
        # app.add_handlers(
        # app.default_router.add_rules(
        #     [
        #         (r"{}redoc".format(prefix), GetRedocHtmlHandle),
        #         (r"{}swagger".format(prefix), GetSwaggerUiHtmlHandle),
        #         (r"{}openapi.json".format(prefix), OpenApiHandle),
        #     ]
        # )
        #
        # Method 3
        app.wildcard_router.add_rules(
            [
                (r"{}redoc".format(prefix), GetRedocHtmlHandle),
                (r"{}swagger".format(prefix), GetSwaggerUiHtmlHandle),
                (r"{}openapi.json".format(prefix), OpenApiHandle),
            ]
        )
        from tornado.web import AnyMatches, Rule, _ApplicationRouter

        app.default_router = _ApplicationRouter(app, [Rule(AnyMatches(), app.wildcard_router)])


def add_doc_route(
    app: Application,
    scheme: Optional[str] = None,
    open_json_url_only_path: bool = False,
    prefix: str = "",
    pin_code: str = "",
    title: str = "",
    open_api_tag_list: Optional[List[Dict[str, Any]]] = None,
    project_name: str = "",
) -> None:
    AddDocRoute(
        scheme=scheme,
        open_json_url_only_path=open_json_url_only_path,
        prefix=prefix,
        pin_code=pin_code,
        title=title,
        open_api_tag_list=open_api_tag_list,
        project_name=project_name,
        app=app,
    )



# flake8: noqa: F403
from ._func_sig import *  # type: ignore
from ._gen_tip import *  # type: ignore
from ._i18n import *  # type: ignore
from ._lazy_property import *  # type: ignore
from ._pydantic_util import *  # type: ignore
from ._types import *  # type: ignore
from ._util import *  # type: ignore

__all__ = [
    _func_sig.__all__  # type: ignore
    + _gen_tip.__all__  # type: ignore
    + _i18n.__all__  # type: ignore
    + _lazy_property.__all__  # type: ignore
    + _pydantic_util.__all__  # type: ignore
    + _types.__all__  # type: ignore
    + _util.__all__  # type: ignore
]



from typing import Any, Dict, List, Optional, Set
from urllib.parse import urlencode

from starlette.applications import Starlette
from starlette.exceptions import HTTPException
from starlette.requests import Request
from starlette.responses import HTMLResponse, Response
from starlette.routing import Mount, Route

from pait.api_doc.html import get_redoc_html as _get_redoc_html
from pait.api_doc.html import get_swagger_ui_html as _get_swagger_ui_html
from pait.api_doc.open_api import PaitOpenAPI
from pait.app.base.doc_route import AddDocRoute as _AddDocRoute
from pait.app.base.doc_route import DocHtmlRespModel, OpenAPIRespModel
from pait.field import Depends
from pait.model.core import PaitCoreModel
from pait.model.template import TemplateContext

from ._load_app import load_app
from ._pait import Pait

__all__ = ["AddDocRoute", "add_doc_route"]
prefix_set_dict: Dict[Starlette, Set[str]] = {}


class AddDocRoute(_AddDocRoute[Starlette]):
    not_found_exc: Exception = HTTPException(
        status_code=404,
        detail=(
            "The requested URL was not found on the server. If you entered"
            " the URL manually please check your spelling and try again."
        ),
    )

    def _gen_route(self, app: Starlette) -> Any:
        # prefix `/` route group must be behind other route group
        if app not in prefix_set_dict:
            prefix_set: Set[str] = set()
            prefix_set_dict[app] = prefix_set
        else:
            prefix_set = prefix_set_dict[app]
        for prefix in prefix_set:
            if self.prefix.startswith(prefix):
                raise RuntimeError(f"prefix:{prefix} already exists, can use:{self.prefix}")

        prefix_set.add(self.prefix)
        doc_pait = self._get_doc_pait(Pait)

        def _get_open_json_url(request: Request, r_pin_code: str, url_dict: Dict[str, Any]) -> str:
            _scheme: str = self.scheme or request.url.scheme
            request_path: str = "/".join(request.url.path.split("/")[:-1])
            if self.open_json_url_only_path:
                openapi_json_url: str = f"{request_path}/openapi.json"
            else:
                url: str = f"{_scheme}://{request.url.hostname}"
                if request.url.port:
                    url += f":{request.url.port}"  # pragma: no cover
                openapi_json_url = f"{url}{request_path}/openapi.json"
            if r_pin_code:
                url_dict["pin_code"] = r_pin_code
            openapi_json_url += "?" + urlencode(url_dict)
            return openapi_json_url

        @doc_pait(response_model_list=[DocHtmlRespModel])
        def get_redoc_html(
            request: Request,
            r_pin_code: str = Depends.i(self._get_request_pin_code),
            url_dict: Dict[str, Any] = Depends.i(self._get_request_template_map()),
        ) -> HTMLResponse:

            return HTMLResponse(
                _get_redoc_html(
                    _get_open_json_url(request, r_pin_code, url_dict),
                    src_url=self.redoc_src_url,
                    title=self.title,
                )
            )

        @doc_pait(response_model_list=[DocHtmlRespModel])
        def get_swagger_ui_html(
            request: Request,
            r_pin_code: str = Depends.i(self._get_request_pin_code),
            url_dict: Dict[str, Any] = Depends.i(self._get_request_template_map()),
        ) -> HTMLResponse:
            return HTMLResponse(
                _get_swagger_ui_html(
                    _get_open_json_url(request, r_pin_code, url_dict),
                    title=self.title,
                    swagger_ui_bundle=self.swagger_ui_bundle,
                    swagger_ui_standalone_preset=self.swagger_ui_standalone_preset,
                    swagger_ui_url=self.swagger_ui_url,
                )
            )

        @doc_pait(pre_depend_list=[self._get_request_pin_code], response_model_list=[OpenAPIRespModel])
        def openapi_route(
            request: Request, url_dict: Dict[str, Any] = Depends.i(self._get_request_template_map(extra_key=True))
        ) -> Response:
            pait_dict: Dict[str, PaitCoreModel] = load_app(request.app, project_name=self.project_name)
            _scheme: str = self.scheme or request.url.scheme
            url: str = f"{_scheme}://{request.url.hostname}"
            if request.url.port:
                url += f":{request.url.port}"  # pragma: no cover
            with TemplateContext(url_dict):
                pait_openapi: PaitOpenAPI = PaitOpenAPI(
                    pait_dict,
                    title=self.title,
                    open_api_server_list=[{"url": url, "description": ""}],
                    open_api_tag_list=self.open_api_tag_list,
                )
                return Response(pait_openapi.content, media_type="application/json")

        route: Mount = Mount(
            self.prefix,
            name=self.title,
            routes=[
                Route("/redoc", get_redoc_html, methods=["GET"]),
                Route("/swagger", get_swagger_ui_html, methods=["GET"]),
                Route("/openapi.json", openapi_route, methods=["GET"]),
            ],
        )
        app.routes.append(route)


def add_doc_route(
    app: Starlette,
    scheme: Optional[str] = None,
    open_json_url_only_path: bool = False,
    prefix: str = "",
    pin_code: str = "",
    title: str = "",
    open_api_tag_list: Optional[List[Dict[str, Any]]] = None,
    project_name: str = "",
) -> None:
    AddDocRoute(
        scheme=scheme,
        open_json_url_only_path=open_json_url_only_path,
        prefix=prefix,
        pin_code=pin_code,
        title=title,
        open_api_tag_list=open_api_tag_list,
        project_name=project_name,
        app=app,
    )



import copy
from typing import TYPE_CHECKING, Any, Dict, Type

from pait.model.response import PaitBaseResponseModel, PaitJsonResponseModel
from pait.plugin.base import PluginProtocol
from pait.util import get_pait_response_model

if TYPE_CHECKING:
    from pait.model.core import PaitCoreModel


class AutoCompleteJsonRespPlugin(PluginProtocol):
    default_response_dict: dict

    def _merge(self, source_dict: dict, target_dict: dict) -> None:
        for key, value in source_dict.items():
            if isinstance(value, dict) and key in target_dict:
                self._merge(value, target_dict[key])
            elif isinstance(value, list) and key in target_dict:
                raw_value = value.pop()
                for item in target_dict[key]:
                    new_value = copy.deepcopy(raw_value)
                    self._merge(new_value, item)
                    value.append(new_value)
            else:
                source_dict[key] = target_dict.get(key, value)

    def merge(self, response_dict: dict) -> dict:
        default_response_dict: dict = copy.deepcopy(self.default_response_dict)
        self._merge(default_response_dict, response_dict)
        return default_response_dict

    @classmethod
    def pre_check_hook(cls, pait_core_model: "PaitCoreModel", kwargs: Dict) -> None:
        super().pre_check_hook(pait_core_model, kwargs)
        if "pait_response_model" in kwargs:
            raise RuntimeError("Please use response_model_list param")

    @classmethod
    def pre_load_hook(cls, pait_core_model: "PaitCoreModel", kwargs: Dict) -> Dict:
        kwargs = super().pre_load_hook(pait_core_model, kwargs)
        pait_response_model: Type[PaitBaseResponseModel] = get_pait_response_model(
            pait_core_model.response_model_list, find_core_response_model=True
        )
        if not issubclass(pait_response_model, PaitJsonResponseModel):
            raise ValueError(f"pait_response_model must {PaitJsonResponseModel} not {pait_response_model}")
        kwargs["default_response_dict"] = pait_response_model.get_default_dict()
        return kwargs

    def _sync_call(self, *args: Any, **kwargs: Any) -> Any:
        response_dict = self.call_next(*args, **kwargs)
        return self.merge(response_dict)

    async def _async_call(self, *args: Any, **kwargs: Any) -> Any:
        response_dict: dict = await self.call_next(*args, **kwargs)
        return self.merge(response_dict)

    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        if self._is_async_func:
            return self._async_call(*args, **kwargs)
        else:
            return self._sync_call(*args, **kwargs)


class AsyncAutoCompleteJsonRespPlugin(AutoCompleteJsonRespPlugin):
    """"""



from typing import TYPE_CHECKING, Callable, List, Optional, Set, Type

from pydantic import BaseConfig, BaseModel

from pait.model.response import PaitBaseResponseModel
from pait.plugin.base import PluginManager
from pait.util import http_method_tuple

if TYPE_CHECKING:
    from pait.model.config import APPLY_FN
    from pait.model.core import MatchRule, PaitCoreModel
    from pait.param_handle import BaseParamHandler


__all__ = [
    "apply_multi_plugin",
    "apply_extra_openapi_model",
    "apply_response_model",
    "apply_default_pydantic_model_config",
    "apply_block_http_method_set",
    "apply_pre_depend",
]


def apply_extra_openapi_model(
    extra_openapi_model: Type[BaseModel], match_rule: Optional["MatchRule"] = None
) -> "APPLY_FN":
    """
    Add a default extre_openapi structure for routing handles
    """

    def _apply(pait_core_model: "PaitCoreModel") -> None:
        if pait_core_model.match(match_rule):
            pait_core_model.extra_openapi_model_list = [extra_openapi_model]

    return _apply


def apply_response_model(
    response_model_list: List[Type[PaitBaseResponseModel]], match_rule: Optional["MatchRule"] = None
) -> "APPLY_FN":
    """
    Add a default response structure for routing handles
    """
    for response_model in response_model_list:
        if response_model.is_core:
            raise ValueError(f"{response_model} is core response model can not set to default_response_model_list")

    def _apply(pait_core_model: "PaitCoreModel") -> None:
        if pait_core_model.match(match_rule):
            pait_core_model.add_response_model_list(response_model_list)

    return _apply


def apply_default_pydantic_model_config(
    pydantic_model_config: Type[BaseConfig], match_rule: Optional["MatchRule"] = None
) -> "APPLY_FN":
    """pait route gen pydantic model default config"""

    def _apply(pait_core_model: "PaitCoreModel") -> None:
        if pait_core_model.match(match_rule):
            pait_core_model.pydantic_model_config = pydantic_model_config

    return _apply


def apply_block_http_method_set(
    block_http_method_set: Set[str], match_rule: Optional["MatchRule"] = None
) -> "APPLY_FN":
    """
    Under normal circumstances, pait.load_app can obtain the http method of the routing handle.
     However, some application frameworks such as flask will automatically add optional http methods
     to the handle, which is great, but you may not want to use them in pait, and pait will not
     automatically recognize them, so you can use parameters to disable certain http method
    """
    for block_http_method in block_http_method_set:
        if block_http_method.lower() not in http_method_tuple:
            raise ValueError(f"Error http method: {block_http_method}")

    def _apply(pait_core_model: "PaitCoreModel") -> None:
        if pait_core_model.match(match_rule):
            pait_core_model.block_http_method_set = block_http_method_set
            pait_core_model.method_list = pait_core_model.method_list

    return _apply


def apply_multi_plugin(
    plugin_manager_fn_list: List[Callable[[], PluginManager]], match_rule: Optional["MatchRule"] = None
) -> "APPLY_FN":
    def _apply(pait_core_model: "PaitCoreModel") -> None:
        if pait_core_model.match(match_rule):
            pre_plugin_manager_list: List[PluginManager] = []
            post_plugin_manager_list: List[PluginManager] = []
            for plugin_manager_fn in plugin_manager_fn_list:
                plugin_manager: PluginManager = plugin_manager_fn()
                if plugin_manager.plugin_class.is_pre_core:
                    pre_plugin_manager_list.append(plugin_manager)
                else:
                    post_plugin_manager_list.append(plugin_manager)
            pait_core_model.add_plugin(pre_plugin_manager_list, post_plugin_manager_list)

    return _apply


def apply_pre_depend(pre_depend: Callable, match_rule: Optional["MatchRule"] = None) -> "APPLY_FN":
    def _apply(pait_core_model: "PaitCoreModel") -> None:
        if pait_core_model.match(match_rule):
            pait_core_model.pre_depend_list.append(pre_depend)

    return _apply


def apply_param_handler(
    param_handler_plugin: "Type[BaseParamHandler]", match_rule: Optional["MatchRule"] = None
) -> "APPLY_FN":
    def _apply(pait_core_model: "PaitCoreModel") -> None:
        if pait_core_model.match(match_rule):
            pait_core_model.param_handler_plugin = param_handler_plugin

    return _apply



from typing import Any, List

from pait.types import Literal

SupportAppLiteral = Literal["flask", "starlette", "sanic", "tornado"]
support_app_list: List[SupportAppLiteral] = ["flask", "starlette", "sanic", "tornado"]


def sniffing(app: Any) -> SupportAppLiteral:
    app_name: str = app.__class__.__name__.lower()
    if app_name in support_app_list:
        return app_name  # type: ignore
    elif app_name == "application" and app.__class__.__module__ == "tornado.web":
        return "tornado"
    else:
        raise NotImplementedError(f"Pait not support app name:{app_name}, please check app:{app}")



from dataclasses import MISSING
from typing import Any, Coroutine, Dict, List, Mapping, Optional

from starlette.datastructures import FormData, Headers, UploadFile
from starlette.requests import Request

from pait.app.base import BaseAppHelper
from pait.util import LazyProperty

__all__ = ["AppHelper"]


class AppHelper(BaseAppHelper):
    RequestType = Request
    FormType = FormData
    FileType = UploadFile
    HeaderType = Headers
    app_name = "starlette"

    def __init__(self, class_: Any, args: List[Any], kwargs: Mapping[str, Any]):
        super().__init__(class_, args, kwargs)
        self._form: Optional[FormData] = None

    def get_attributes(self, key: str, default: Any = MISSING) -> Any:
        if default is MISSING:
            return getattr(self.request.app.state, key)
        else:
            return getattr(self.request.app.state, key, default)

    def body(self) -> dict:
        return self.request.json()

    def cookie(self) -> dict:
        return self.request.cookies

    async def get_form(self) -> FormData:
        if self._form:
            return self._form
        form: FormData = await self.request.form()
        self._form = form
        return form

    def file(self) -> Coroutine[Any, Any, FormData]:
        async def _() -> FormData:
            return await self.get_form()

        return _()

    def form(self) -> Coroutine[Any, Any, Dict[str, Any]]:
        @LazyProperty()
        async def _form() -> Dict[str, Any]:
            form_data: FormData = await self.get_form()
            return {key: form_data.getlist(key)[0] for key, _ in form_data.items()}

        return _form()

    def header(self) -> Headers:
        return self.request.headers

    def path(self) -> dict:
        return self.request.path_params

    def query(self) -> dict:
        return dict(self.request.query_params)

    def multiform(self) -> Coroutine[Any, Any, Dict[str, List[Any]]]:
        @LazyProperty()
        async def _multiform() -> Dict[str, List[Any]]:
            form_data: FormData = await self.get_form()
            return {
                key: [i for i in form_data.getlist(key) if not isinstance(i, UploadFile)]
                for key, _ in form_data.items()
            }

        return _multiform()

    @LazyProperty()
    def multiquery(self) -> Dict[str, Any]:
        return {key: self.request.query_params.getlist(key) for key, _ in self.request.query_params.items()}



from typing import Any

from pait.plugin.auto_complete_json_resp import AutoCompleteJsonRespPlugin as _AutoCompleteJsonRespPlugin

from .base import JsonProtocol

__all__ = ["AutoCompleteJsonRespPlugin", "AsyncAutoCompleteJsonRespPlugin"]


class AutoCompleteJsonRespPlugin(JsonProtocol, _AutoCompleteJsonRespPlugin):
    def _sync_call(self, *args: Any, **kwargs: Any) -> Any:
        response: Any = super()._sync_call(*args, **kwargs)
        return self.gen_response(response)

    async def _async_call(self, *args: Any, **kwargs: Any) -> Any:
        response: Any = await super()._async_call(*args, **kwargs)
        return self.gen_response(response)


class AsyncAutoCompleteJsonRespPlugin(AutoCompleteJsonRespPlugin):
    """"""



import inspect
import logging
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, Callable, List, Optional, Set, Tuple, Type

from pydantic import BaseConfig, BaseModel

from pait.model.response import PaitBaseResponseModel, PaitResponseModel
from pait.model.status import PaitStatus
from pait.param_handle import AsyncParamHandler, BaseParamHandler, ParamHandler
from pait.plugin import PluginManager
from pait.types import Literal
from pait.util import ignore_pre_check

if TYPE_CHECKING:
    from pait.app.base import BaseAppHelper


__all__ = ["PaitCoreModel", "MatchRule", "MatchKeyLiteral", "ContextModel"]
MatchKeyLiteral = Literal[
    "all",
    "status",
    "group",
    "tag",
    "method_list",
    "path",
    "!status",
    "!group",
    "!tag",
    "!method_list",
    "!path",
]


@dataclass
class MatchRule(object):
    key: MatchKeyLiteral = "all"
    target: Any = None


@dataclass
class ContextModel(object):
    cbv_instance: Optional[Any]
    app_helper: "BaseAppHelper"


class PaitCoreModel(object):
    _param_handler_plugin: PluginManager["BaseParamHandler"]

    def __init__(
        self,
        func: Callable,
        app_helper_class: "Type[BaseAppHelper]",
        pre_depend_list: Optional[List[Callable]] = None,
        path: Optional[str] = None,
        openapi_path: Optional[str] = None,
        method_set: Optional[Set[str]] = None,
        operation_id: Optional[str] = None,
        func_name: Optional[str] = None,
        author: Optional[Tuple[str, ...]] = None,
        summary: Optional[str] = None,
        desc: Optional[str] = None,
        status: Optional[PaitStatus] = None,
        group: Optional[str] = None,
        tag: Optional[Tuple[str, ...]] = None,
        response_model_list: Optional[List[Type[PaitBaseResponseModel]]] = None,
        pydantic_model_config: Optional[Type[BaseConfig]] = None,
        plugin_list: Optional[List[PluginManager]] = None,
        post_plugin_list: Optional[List[PluginManager]] = None,
        param_handler_plugin: Optional[Type[BaseParamHandler]] = None,
        feature_code: str = "",
    ):
        # pait
        self.app_helper_class: "Type[BaseAppHelper]" = app_helper_class
        self.func: Callable = func  # route func
        # self.qualname: str = func.__qualname__.split(".<locals>", 1)[0].rsplit(".", 1)[0]
        self.pait_id: str = f"{func.__qualname__}_{self.func_md5}"
        # Some functions have the same md5 as the name and need to be distinguished by the feature code
        if feature_code:
            self.pait_id = f"{feature_code}_{self.pait_id}"
        setattr(func, "_pait_id", self.pait_id)
        setattr(func, "pait_core_model", self)
        self.pre_depend_list: List[Callable] = pre_depend_list or []
        self.func_path: str = ""
        self.block_http_method_set: Set[str] = set()
        self.pydantic_model_config: Type[BaseConfig] = pydantic_model_config or BaseConfig

        # api doc
        self.path: str = path or ""  # request url path
        self.openapi_path: str = openapi_path or ""
        self._method_list: List[str] = sorted(list(method_set or set()))  # request method set
        self.func_name: str = func_name or func.__name__
        self.operation_id: str = operation_id or self.func_name  # route name
        self.author: Optional[Tuple[str, ...]] = author  # The main developer of this func
        self.summary: str = summary or ""
        self.desc: str = desc or func.__doc__ or ""  # desc of this func
        self.status: PaitStatus = status or PaitStatus.undefined
        self.group: str = group or "root"  # Which group this interface belongs to
        self.tag: Tuple[str, ...] = tag or ("default",)  # Interface tag
        self._extra_openapi_model_list: List[Type[BaseModel]] = []

        self._response_model_list: List[Type[PaitBaseResponseModel]] = []
        if response_model_list:
            self.add_response_model_list(response_model_list)

        # pait plugin
        self._plugin_list: List[PluginManager] = []
        self._post_plugin_list: List[PluginManager] = []
        self._plugin_manager_list: List[PluginManager] = []

        self.param_handler_plugin = param_handler_plugin  # type: ignore
        self.add_plugin(plugin_list, post_plugin_list)

    @property
    def param_handler_plugin(self) -> Type[BaseParamHandler]:
        return self._param_handler_plugin.plugin_class

    @param_handler_plugin.setter
    def param_handler_plugin(self, param_handler_plugin: Optional[Type[BaseParamHandler]]) -> None:
        if param_handler_plugin:
            self._param_handler_plugin = PluginManager(param_handler_plugin)
        elif inspect.iscoroutinefunction(self.func):
            self._param_handler_plugin = PluginManager(AsyncParamHandler)
        else:
            self._param_handler_plugin = PluginManager(ParamHandler)
        if not ignore_pre_check:
            self._param_handler_plugin.pre_check_hook(self)
        self._param_handler_plugin.pre_load_hook(self)
        if not self._plugin_manager_list:
            self.add_plugin([], [])

    @property
    def func_md5(self) -> str:
        from hashlib import md5

        h = md5()
        h.update(self.func.__code__.co_code)  # type: ignore
        return h.hexdigest()

    @property
    def method_list(self) -> List[str]:
        _temp_set: Set[str] = set(self._method_list.copy())
        _temp_set.difference_update(self.block_http_method_set)
        return sorted(list(_temp_set))

    @method_list.setter
    def method_list(self, method_list: List[str]) -> None:
        self._method_list = list(set(self._method_list) | set(method_list))

    @property
    def response_model_list(self) -> List[Type[PaitBaseResponseModel]]:
        return self._response_model_list

    def add_response_model_list(self, response_model_list: List[Type[PaitBaseResponseModel]]) -> None:
        for response_model in response_model_list:
            if response_model in self._response_model_list:
                continue
            if issubclass(response_model, PaitResponseModel):
                logging.warning(
                    f"Please replace {self.operation_id}'s response model {response_model}"
                    f" with {PaitBaseResponseModel}"
                )
            self._response_model_list.append(response_model)

    @property
    def extra_openapi_model_list(self) -> List[Type[BaseModel]]:
        return self._extra_openapi_model_list

    @extra_openapi_model_list.setter
    def extra_openapi_model_list(self, item: List[Type[BaseModel]]) -> None:
        self._extra_openapi_model_list.extend(item)

    @property
    def plugin_list(self) -> List[PluginManager]:
        return self._plugin_manager_list

    def match(self, match_rule: Optional[MatchRule] = None) -> bool:
        """By different attributes to determine whether to match with pait_core_model,
        if the key is `all` then match
        if the key is prefixed with ! then the result will be reversed
        """
        if not match_rule:
            match_rule = MatchRule()

        key: MatchKeyLiteral = match_rule.key
        target: Any = match_rule.target
        if key == "all":
            return True
        is_reverse: bool = False
        if key.startswith("!"):
            key = key[1:]  # type: ignore
            is_reverse = True

        value: Any = getattr(self, key, ...)
        if value is ...:
            raise KeyError(f"match fail, not found key:{key}")
        if key in ("status", "group"):
            result: bool = value is target
        elif key in ("tag", "method_list"):
            result = target in value
        elif key == "path":
            result = value.startswith(target)
        else:
            raise KeyError(f"Not support key:{key}")

        if is_reverse:
            return not result
        else:
            return result

    def add_plugin(
        self, plugin_list: Optional[List[PluginManager]], post_plugin_list: Optional[List[PluginManager]]
    ) -> None:
        raw_plugin_list: List[PluginManager] = self._plugin_list
        raw_post_plugin_list: List[PluginManager] = self._post_plugin_list
        try:
            for plugin_manager in plugin_list or []:
                if not plugin_manager.plugin_class.is_pre_core:
                    raise ValueError(f"{plugin_manager.plugin_class} is post plugin")
                if not ignore_pre_check:
                    plugin_manager.pre_check_hook(self)
                plugin_manager.pre_load_hook(self)
                self._plugin_list.append(plugin_manager)

            for plugin_manager in post_plugin_list or []:
                if plugin_manager.plugin_class.is_pre_core:
                    raise ValueError(f"{plugin_manager.plugin_class} is pre plugin")
                if not ignore_pre_check:
                    plugin_manager.pre_check_hook(self)
                plugin_manager.pre_load_hook(self)
                self._post_plugin_list.append(plugin_manager)
        except Exception as e:
            self._plugin_list = raw_plugin_list
            self._post_plugin_list = raw_post_plugin_list
            raise e
        else:
            # In future version, it may be possible to switch plugins at runtime
            plugin_manager_list: List[PluginManager] = (
                [i for i in self._plugin_list] + [self._param_handler_plugin] + [i for i in self._post_plugin_list]
            )
            # copy.deepcopy(
            #     self._plugin_list + [self._param_handler_plugin] + self._post_plugin_list
            # )
            plugin_manager_list.reverse()
            self._plugin_manager_list = plugin_manager_list



import inspect
import logging
from typing import TYPE_CHECKING, Any, Dict, Generic, Type, TypeVar

if TYPE_CHECKING:
    from pait.model.core import PaitCoreModel

_PluginT = TypeVar("_PluginT", bound="PluginProtocol")
logger: logging.Logger = logging.getLogger(__name__)


class PluginProtocol(object):
    is_pre_core: bool = True
    pait_core_model: "PaitCoreModel"
    args: list
    kwargs: dict

    _is_async_func: bool

    def __init__(self: "_PluginT", **kwargs: Any) -> None:
        """Direct init calls are not supported,
        so there is no need to write clearly in init what parameters are needed
        """
        if kwargs:
            for k, v in kwargs.items():
                if getattr(self, k, None) is not None:
                    continue
                setattr(self, k, v)

    @classmethod
    def pre_check_hook(cls, pait_core_model: "PaitCoreModel", kwargs: Dict) -> None:
        class_name: str = cls.__class__.__name__
        if class_name.startswith("Async"):
            logger.warning(
                f"Please use {class_name.replace('Async', '')}, {class_name} will remove on version 1.0"
            )  # pragma: no cover

    @classmethod
    def pre_load_hook(cls, pait_core_model: "PaitCoreModel", kwargs: Dict) -> Dict:
        kwargs["_is_async_func"] = inspect.iscoroutinefunction(pait_core_model.func)
        return kwargs  # pragma: no cover

    def __post_init__(self, pait_core_model: "PaitCoreModel", args: tuple, kwargs: dict) -> None:
        self.pait_core_model = pait_core_model
        self.args = list(args) or []
        self.kwargs = kwargs or {}

    @classmethod
    def build(cls, **kwargs: Any) -> "PluginManager[_PluginT]":
        return PluginManager(cls, **kwargs)  # type: ignore

    def call_next(self, *args: Any, **kwargs: Any) -> Any:
        raise RuntimeError("Failed to load Plugin, please check the list of plugins")  # pragma: no cover

    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        return self.call_next(*args, **kwargs)  # pragma: no cover


class PluginManager(Generic[_PluginT]):
    def __init__(self, plugin_class: Type[_PluginT], **kwargs: Any):
        self.plugin_class: Type[_PluginT] = plugin_class
        self._kwargs: Any = kwargs

    def pre_check_hook(self, pait_core_model: "PaitCoreModel") -> None:
        self.plugin_class.pre_check_hook(pait_core_model, self._kwargs)

    def pre_load_hook(self, pait_core_model: "PaitCoreModel") -> None:
        self._kwargs = self.plugin_class.pre_load_hook(pait_core_model, self._kwargs)

    def get_plugin(self) -> _PluginT:
        return self.plugin_class(**self._kwargs)

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} {self.plugin_class.__name__}>"



import inspect
from dataclasses import dataclass
from typing import Callable, Dict, List, Optional, Type

from typing_extensions import Self  # type: ignore

from ._util import get_real_annotation

__all__ = ["FuncSig", "get_func_sig", "is_bounded_func"]


@dataclass()
class FuncSig:
    """func inspect.Signature model"""

    func: Callable
    sig: "inspect.Signature"
    param_list: List["inspect.Parameter"]
    cbv_class: Optional[Type] = None


_func_sig_dict: Dict[Callable, FuncSig] = {}


def get_func_sig(func: Callable) -> FuncSig:
    """get func inspect.Signature model"""
    if func in _func_sig_dict:
        return _func_sig_dict[func]

    sig: inspect.Signature = inspect.signature(func)
    param_list: List[inspect.Parameter] = []
    for key in sig.parameters:
        # NOTE:
        #   The cbv func decorated in Pait is unbound, so it can get to self；
        #   Depend func must be bound func when it is used, so it cannot get self；
        if not (sig.parameters[key].annotation != sig.empty or sig.parameters[key].name == "self"):
            if sig.parameters[key].annotation != Self:
                continue
        parameter: inspect.Parameter = sig.parameters[key]
        setattr(parameter, "_annotation", get_real_annotation(parameter.annotation, func))
        param_list.append(parameter)

    # return_param = sig.return_annotation
    func_sig: FuncSig = FuncSig(func=func, sig=sig, param_list=param_list)
    _func_sig_dict[func] = func_sig
    return func_sig


def is_bounded_func(func: Callable) -> bool:
    return inspect.signature(func).parameters.get("self", None) is None



from typing import Any, Callable

from flask import Blueprint, Flask, Response, jsonify

from pait.app.base.grpc_route import GrpcGatewayRoute as BaseGrpcRouter
from pait.app.flask import pait


def make_response(_: Any, resp_dict: dict) -> Response:
    return jsonify(resp_dict)


class GrpcGatewayRoute(BaseGrpcRouter):
    pait = pait
    make_response: Callable = make_response

    def _add_route(self, app: Flask) -> Any:
        for parse_stub in self.parse_stub_list:
            blueprint: Blueprint = Blueprint(self.title + parse_stub.name, __name__, url_prefix=self.prefix)
            for method_name, grpc_model in parse_stub.method_dict.items():
                _route, grpc_pait_model = self._gen_route_func(method_name, grpc_model)
                if not _route:
                    continue
                # grpc http method only POST
                blueprint.add_url_rule(self.url_handler(grpc_pait_model.url), view_func=_route, methods=["POST"])
            app.register_blueprint(blueprint)



from typing import Any, Dict, List, Optional
from urllib.parse import urlencode

from flask import Blueprint, Flask, Response, current_app, make_response, request
from werkzeug.exceptions import NotFound

from pait.api_doc.html import get_redoc_html as _get_redoc_html
from pait.api_doc.html import get_swagger_ui_html as _get_swagger_ui_html
from pait.api_doc.open_api import PaitOpenAPI
from pait.app.base.doc_route import AddDocRoute as _AddDocRoute
from pait.app.base.doc_route import DocHtmlRespModel, OpenAPIRespModel
from pait.field import Depends
from pait.model.core import PaitCoreModel
from pait.model.template import TemplateContext

from ._load_app import load_app
from ._pait import Pait

__all__ = ["add_doc_route", "AddDocRoute"]


class AddDocRoute(_AddDocRoute[Flask]):
    not_found_exc: Exception = NotFound()

    def _gen_route(self, app: Flask) -> Any:
        doc_pait = self._get_doc_pait(Pait)

        def _get_open_json_url(r_pin_code: str, url_dict: Dict[str, Any]) -> str:
            _scheme: str = self.scheme or request.scheme
            request_path: str = "/".join(request.path.split("/")[:-1])
            if self.open_json_url_only_path:
                openapi_json_url: str = f"{request_path}/openapi.json"
            else:
                openapi_json_url = f"{_scheme}://{request.host}{request_path}/openapi.json"
            if r_pin_code:
                url_dict["pin_code"] = r_pin_code
            openapi_json_url += "?" + urlencode(url_dict)
            return openapi_json_url

        @doc_pait(response_model_list=[DocHtmlRespModel])
        def get_redoc_html(
            r_pin_code: str = Depends.i(self._get_request_pin_code),
            url_dict: Dict[str, Any] = Depends.i(self._get_request_template_map()),
        ) -> str:
            return _get_redoc_html(
                _get_open_json_url(r_pin_code, url_dict),
                src_url=self.redoc_src_url,
                title=self.title,
            )

        @doc_pait(response_model_list=[DocHtmlRespModel])
        def get_swagger_ui_html(
            r_pin_code: str = Depends.i(self._get_request_pin_code),
            url_dict: Dict[str, Any] = Depends.i(self._get_request_template_map()),
        ) -> str:
            return _get_swagger_ui_html(
                _get_open_json_url(r_pin_code, url_dict),
                title=self.title,
                swagger_ui_bundle=self.swagger_ui_bundle,
                swagger_ui_standalone_preset=self.swagger_ui_standalone_preset,
                swagger_ui_url=self.swagger_ui_url,
            )

        @doc_pait(pre_depend_list=[self._get_request_pin_code], response_model_list=[OpenAPIRespModel])
        def openapi_route(
            url_dict: Dict[str, Any] = Depends.i(self._get_request_template_map(extra_key=True))
        ) -> Response:
            _scheme: str = self.scheme or request.scheme
            pait_dict: Dict[str, PaitCoreModel] = load_app(current_app, project_name=self.project_name)
            with TemplateContext(url_dict):
                pait_openapi: PaitOpenAPI = PaitOpenAPI(
                    pait_dict,
                    title=self.title,
                    open_api_server_list=[{"url": f"{_scheme}://{request.host}", "description": ""}],
                    open_api_tag_list=self.open_api_tag_list,
                )
                response: Response = make_response(pait_openapi.content)
                response.mimetype = "application/json"
            return response

        blueprint: Blueprint = Blueprint(self.title, __name__, url_prefix=self.prefix)
        blueprint.add_url_rule("/redoc", view_func=get_redoc_html, methods=["GET"])
        blueprint.add_url_rule("/swagger", view_func=get_swagger_ui_html, methods=["GET"])
        blueprint.add_url_rule("/openapi.json", view_func=openapi_route, methods=["GET"])
        app.register_blueprint(blueprint)


def add_doc_route(
    app: Flask,
    scheme: Optional[str] = None,
    open_json_url_only_path: bool = False,
    prefix: str = "",
    pin_code: str = "",
    title: str = "",
    open_api_tag_list: Optional[List[Dict[str, Any]]] = None,
    project_name: str = "",
) -> None:
    AddDocRoute(
        scheme=scheme,
        open_json_url_only_path=open_json_url_only_path,
        prefix=prefix,
        pin_code=pin_code,
        title=title,
        open_api_tag_list=open_api_tag_list,
        project_name=project_name,
        app=app,
    )



from typing import Any, Dict, Optional

from tornado.web import RequestHandler

from pait.plugin.base import PluginManager, PluginProtocol


class JsonProtocol(PluginProtocol):
    status: int
    headers: Dict[str, str]
    content_type: str

    def gen_response(self, tornado_handle: RequestHandler, response: Any) -> Any:
        tornado_handle.set_status(self.status)
        if self.headers is not None:
            for k, v in self.headers.items():
                tornado_handle.set_header(k, v)
        tornado_handle.write(response)
        return response

    @classmethod
    def build(  # type: ignore
        cls,
        status: Optional[int] = None,
        headers: Optional[Dict[str, str]] = None,
        content_type: Optional[str] = None,
        **kwargs: Any,
    ) -> "PluginManager":
        return super().build(
            status=status or 200,
            headers=headers or {},
            content_type=content_type or "application/json",
            **kwargs,
        )



import pickle
from typing import TYPE_CHECKING, Any, Callable, Dict, Optional, Tuple, Type, Union

from redis import Redis  # type: ignore
from redis.asyncio import Redis as AsyncioRedis  # type: ignore

from pait.app import set_app_attribute
from pait.g import pait_context
from pait.model.response import PaitFileResponseModel
from pait.plugin.base import PluginProtocol

if TYPE_CHECKING:
    from pait.model.core import PaitCoreModel
    from pait.plugin.base import PluginManager


_cache_plugin_redis_key: str = "_cache_plugin_redis"


class CacheResponsePlugin(PluginProtocol):
    is_pre_core: bool = False
    name: str
    lock_name: str
    include_exc: Optional[Tuple[Type[Exception]]] = None
    redis: Union[Redis, AsyncioRedis, None] = None
    enable_cache_name_merge_param: bool
    cache_time: Optional[int]
    timeout: Optional[float]
    sleep: Optional[float]
    blocking_timeout: Optional[float]

    def __post_init__(self, pait_core_model: "PaitCoreModel", args: tuple, kwargs: dict) -> None:
        self.lock_name: str = self.name + ":" + "lock"

    @classmethod
    def check_redis(cls, redis: Union[Redis, AsyncioRedis]) -> None:
        if redis.connection_pool.connection_kwargs["decode_responses"] is False:
            raise ValueError("Please set redis`s param:decode_responses to True")

    @classmethod
    def set_redis_to_app(cls, app: Any, redis: Union[Redis, AsyncioRedis]) -> None:
        cls.check_redis(redis)
        set_app_attribute(app, _cache_plugin_redis_key, redis)

    @classmethod
    def pre_check_hook(cls, pait_core_model: "PaitCoreModel", kwargs: Dict) -> None:
        super().pre_check_hook(pait_core_model, kwargs)
        if not pait_core_model.response_model_list:
            raise RuntimeError(f"{pait_core_model.func} can not found response model")
        if issubclass(pait_core_model.response_model_list[0], PaitFileResponseModel):
            raise RuntimeError(
                f"Not use {cls.__name__} in {pait_core_model.func.__name__}, "
                f"{cls.__name__} not support {PaitFileResponseModel.__class__.__name__}"
            )
        if "pait_response_model" in kwargs:
            raise RuntimeError("Please use response_model_list param")  # pragma: no cover

        if kwargs.get("redis", None) is not None:
            cls.check_redis(kwargs["redis"])

    @classmethod
    def pre_load_hook(cls, pait_core_model: "PaitCoreModel", kwargs: Dict) -> Dict:
        super().pre_load_hook(pait_core_model, kwargs)
        name: str = kwargs.get("name", "")
        if not name:
            kwargs["name"] = pait_core_model.func.__qualname__
        kwargs["pait_response_model"] = pait_core_model.response_model_list[0]
        return kwargs

    def _get_redis(self) -> Union[Redis, AsyncioRedis]:
        redis: Union[Redis, AsyncioRedis, None] = self.redis or pait_context.get().app_helper.get_attributes(
            _cache_plugin_redis_key, None
        )
        if not redis:
            raise ValueError("Not found redis client")
        return redis

    def _gen_key(self, *args: Any, **kwargs: Any) -> Tuple[str, str]:
        real_key: str = self.name
        real_lock_key: str = self.lock_name
        if self.enable_cache_name_merge_param:
            args_key_list: list = []
            if args:
                args_key_list = [str(i) for i in args]
            if kwargs:
                for value in kwargs.values():
                    args_key_list.append(str(value))
            if args_key_list:
                real_key = f"{self.name}:{':'.join(args_key_list)}"
                real_lock_key = f"{self.lock_name}:{':'.join(args_key_list)}"
        return real_key, real_lock_key

    def _loads(self, response: Any, *args: Any, **kwargs: Any) -> Any:
        return pickle.loads(response.encode("latin1"))

    def _dumps(self, response: Any, *args: Any, **kwargs: Any) -> Any:
        return pickle.dumps(response).decode("latin1")

    async def _async_cache(self, func: Callable, *args: Any, **kwargs: Any) -> Any:
        real_key, real_lock_key = self._gen_key(*args, **kwargs)
        redis: AsyncioRedis = self._get_redis()
        result: Any = await redis.get(real_key)
        if result:
            result = self._loads(result, *args, **kwargs)
        else:
            async with redis.lock(
                real_lock_key,
                timeout=self.timeout,
                sleep=self.sleep,
                blocking_timeout=self.blocking_timeout,
            ):
                result = await redis.get(real_key)
                if result:
                    result = self._loads(result, *args, **kwargs)
                else:
                    try:
                        result = await func(*args, **kwargs)
                    except Exception as e:
                        if self.include_exc and isinstance(e, self.include_exc):
                            result = e
                        else:
                            raise e
                    await redis.set(real_key, self._dumps(result, *args, **kwargs), ex=self.cache_time)  # type: ignore
        if isinstance(result, Exception):
            raise result
        return result

    def _cache(self, func: Callable, *args: Any, **kwargs: Any) -> Any:
        real_key, real_lock_key = self._gen_key(*args, **kwargs)
        redis: Redis = self._get_redis()
        result: Any = redis.get(real_key)
        if result:
            result = self._loads(result, *args, **kwargs)
        else:
            with redis.lock(
                real_lock_key,
                timeout=self.timeout,
                sleep=self.sleep,
                blocking_timeout=self.blocking_timeout,
            ):
                result = redis.get(real_key)
                if result:
                    result = self._loads(result, *args, **kwargs)
                else:
                    try:
                        result = func(*args, **kwargs)
                    except Exception as e:
                        if self.include_exc and isinstance(e, self.include_exc):
                            result = e
                        else:
                            raise e
                    redis.set(real_key, self._dumps(result, *args, **kwargs), ex=self.cache_time)
        if isinstance(result, Exception):
            raise result
        return result

    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        if self._is_async_func:
            return self._async_cache(self.call_next, *args, **kwargs)
        else:
            return self._cache(self.call_next, *args, **kwargs)

    @classmethod
    def build(  # type: ignore
        cls,
        *,
        redis: Union[Redis, AsyncioRedis, None] = None,
        include_exc: Optional[Tuple[Type[Exception]]] = None,
        name: str = "",
        enable_cache_name_merge_param: bool = False,
        cache_time: Optional[int] = None,
        timeout: Optional[float] = None,
        sleep: Optional[float] = None,
        blocking_timeout: Optional[float] = None,
    ) -> "PluginManager":  # type: ignore
        return super().build(
            name=name,
            redis=redis,
            include_exc=include_exc,
            enable_cache_name_merge_param=enable_cache_name_merge_param,
            cache_time=cache_time or 5 * 60,
            timeout=timeout,
            sleep=sleep,
            blocking_timeout=blocking_timeout,
        )



from typing import Any, Callable, List

from starlette.applications import Starlette
from starlette.responses import JSONResponse, Response
from starlette.routing import Mount, Route

from pait.app.base.grpc_route import AsyncGrpcGatewayRoute as BaseGrpcRouter
from pait.app.starlette import pait


def make_response(_: Any, resp_dict: dict) -> Response:
    return JSONResponse(resp_dict)


class GrpcGatewayRoute(BaseGrpcRouter):
    pait = pait
    make_response: Callable = make_response

    def _add_route(self, app: Starlette) -> Any:
        route_list: List[Route] = []
        for parse_stub in self.parse_stub_list:
            for method_name, grpc_model in parse_stub.method_dict.items():
                _route, grpc_pait_model = self._gen_route_func(method_name, grpc_model)
                if not _route:
                    continue

                # grpc http method only POST
                route_list.append(Route(self.url_handler(grpc_pait_model.url), _route, methods=["POST"]))
        app.routes.append(Mount(self.prefix, name=self.title, routes=route_list))



from google.protobuf.descriptor import Descriptor, FieldDescriptor  # type: ignore
from google.protobuf.json_format import MessageToDict  # type: ignore
from google.protobuf.message import Message  # type: ignore
from google.protobuf.timestamp_pb2 import Timestamp  # type: ignore

__all__ = ["Descriptor", "FieldDescriptor", "Message", "Timestamp", "MessageToDict"]



from typing import Optional


def get_redoc_html(
    open_api_json_url: str,
    src_url: Optional[str] = None,
    title: str = "ReDoc",
) -> str:
    """copy from https://github.com/Redocly/redoc#tldr"""
    if not src_url:
        src_url = "https://cdn.jsdelivr.net/npm/redoc@next/bundles/redoc.standalone.js"
    return f"""
    <!DOCTYPE html>
    <html>
      <head>
        <title>{title}</title>
        <!-- needed for adaptive design -->
        <meta charset="utf-8"/>
        <meta name="viewport" content="width=device-width, initial-scale=1">
        <link href="https://fonts.googleapis.com/css?family=Montserrat:300,400,700|Roboto:300,400,700" rel="stylesheet">

        <!--
        ReDoc doesn't change outer page styles
        -->
        <style>
          body {{
            margin: 0;
            padding: 0;
          }}
        </style>
      </head>
      <body>
        <redoc spec-url='{open_api_json_url}'></redoc>
        <script src={src_url}> </script>
      </body>
    </html>
    """


def get_swagger_ui_html(
    open_api_json_url: str,
    title: str = "Swagger",
    swagger_ui_url: Optional[str] = None,
    swagger_ui_bundle: Optional[str] = None,
    swagger_ui_standalone_preset: Optional[str] = None,
) -> str:
    if not swagger_ui_url:
        swagger_ui_url = "https://unpkg.com/swagger-ui-dist@4.5.0/swagger-ui.css"
    if not swagger_ui_bundle:
        swagger_ui_bundle = "https://unpkg.com/swagger-ui-dist@4.5.0/swagger-ui-bundle.js"
    if not swagger_ui_standalone_preset:
        swagger_ui_standalone_preset = "https://unpkg.com/swagger-ui-dist@4.5.0/swagger-ui-standalone-preset.js"

    return f"""
<!DOCTYPE html>
<html lang="en">
  <head>
    <meta charset="utf-8" />
    <meta name="viewport" content="width=device-width, initial-scale=1" />
    <meta
      name="description"
      content="SwaggerUI"
    />
    <title>{title}</title>
    <link rel="stylesheet" href="{swagger_ui_bundle}" />
  </head>
  <body>
  <div id="swagger-ui"></div>
  <script src="{swagger_ui_bundle}" crossorigin></script>
  <script src="{swagger_ui_standalone_preset}" crossorigin></script>
  <script>
    window.onload = () => {{
      window.ui = SwaggerUIBundle({{
        url: '{open_api_json_url}',
        dom_id: '#swagger-ui',
        presets: [
          SwaggerUIBundle.presets.apis,
          SwaggerUIStandalonePreset
        ],
        layout: "StandaloneLayout",
        deepLinking: true,
        showExtensions: true,
        showCommonExtensions: true
      }});
    }};
  </script>
  </body>
</html>
"""
    # return f"""
    # <!DOCTYPE html>
    # <html>
    # <head>
    # <link type="text/css" rel="stylesheet" href="https://cdn.jsdelivr.net/npm/swagger-ui-dist@3/swagger-ui.css">
    # <link rel="shortcut icon" href="https://static1.smartbear.co/swagger/media/assets/swagger_fav.png">
    # <title>{title}</title>
    # </head>
    # <body>
    # <div id="swagger-ui">
    # </div>
    # <script src="https://cdn.jsdelivr.net/npm/swagger-ui-dist@3/swagger-ui-bundle.js"></script>
    # <!-- `SwaggerUIBundle` is now available on the page -->
    # <script>
    # const ui = SwaggerUIBundle({{
    #     url: '{open_api_json_url}',
    #     dom_id: '#swagger-ui',
    #     presets: [
    #     SwaggerUIBundle.presets.apis,
    #     SwaggerUIBundle.SwaggerUIStandalonePreset
    #     ],
    #     layout: "BaseLayout",
    #     deepLinking: true,
    #     showExtensions: true,
    #     showCommonExtensions: true
    # }})
    # </script>
    # </body>
    # </html>
    # """



import sys

# copy from https://github.com/agronholm/typeguard/blob/master/src/typeguard/__init__.py#L64
if sys.version_info >= (3, 10):
    from typing import ParamSpec  # pragma: no cover
    from typing import is_typeddict  # pragma: no cover
    from typing import Literal
else:
    from typing_extensions import ParamSpec  # type: ignore
    from typing_extensions import Literal

    _typed_dict_meta_types = ()
    if sys.version_info >= (3, 8):
        from typing import _TypedDictMeta  # pragma: no cover

        _typed_dict_meta_types += (_TypedDictMeta,)  # pragma: no cover

    try:
        from typing_extensions import _TypedDictMeta  # type: ignore

        _typed_dict_meta_types += (_TypedDictMeta,)  # type: ignore
    except ImportError:  # pragma: no cover
        pass

    def is_typeddict(tp) -> bool:  # type: ignore
        return isinstance(tp, _typed_dict_meta_types)


__all__ = ["ParamSpec", "Literal", "is_typeddict"]



from pait.app.starlette.plugin.check_json_resp import AsyncCheckJsonRespPlugin, CheckJsonRespPlugin
from pait.app.starlette.plugin.mock_response import AsyncMockPlugin, MockPlugin
from pait.plugin.at_most_one_of import AsyncAtMostOneOfPlugin, AtMostOneOfPlugin
from pait.plugin.required import AsyncRequiredPlugin, RequiredPlugin

__all__ = [
    "RequiredPlugin",
    "AsyncRequiredPlugin",
    "AsyncAtMostOneOfPlugin",
    "AtMostOneOfPlugin",
    "AsyncCheckJsonRespPlugin",
    "AsyncMockPlugin",
    "CheckJsonRespPlugin",
    "MockPlugin",
]



import datetime
import json
from dataclasses import MISSING
from enum import IntEnum
from typing import Any, Callable, Dict, List, Optional, Tuple, Type, Union

from pydantic import BaseModel, Field, validator
from pydantic.fields import FieldInfo, Undefined
from pydantic.typing import NoArgAnyCallable

from pait.field import Depends
from pait.util import create_pydantic_model
from pait.util.grpc_inspect.types import Descriptor, FieldDescriptor, Message, Timestamp

type_dict: Dict[str, Type] = {
    FieldDescriptor.TYPE_DOUBLE: float,
    FieldDescriptor.TYPE_FLOAT: float,
    FieldDescriptor.TYPE_INT64: int,
    FieldDescriptor.TYPE_UINT64: int,
    FieldDescriptor.TYPE_INT32: int,
    FieldDescriptor.TYPE_FIXED64: float,
    FieldDescriptor.TYPE_FIXED32: float,
    FieldDescriptor.TYPE_BOOL: bool,
    FieldDescriptor.TYPE_STRING: str,
    FieldDescriptor.TYPE_BYTES: str,
    FieldDescriptor.TYPE_UINT32: int,
    FieldDescriptor.TYPE_SFIXED32: float,
    FieldDescriptor.TYPE_SFIXED64: float,
    FieldDescriptor.TYPE_SINT32: int,
    FieldDescriptor.TYPE_SINT64: int,
}

GRPC_TIMESTAMP_HANDLER_TUPLE_T = Tuple[Any, Optional[Callable[[Any, Any], Timestamp]]]


class MessagePaitModel(BaseModel):
    enable: bool = Field(True)
    miss_default: bool = Field(False)
    example: Any = Field(MISSING)
    alias: Optional[str] = Field(None)
    title: Optional[str] = Field(None)
    description: Optional[str] = Field(None)
    const: Optional[bool] = Field(None)
    gt: Union[int, float, None] = Field(None)
    ge: Union[int, float, None] = Field(None)
    lt: Union[int, float, None] = Field(None)
    le: Union[int, float, None] = Field(None)
    min_length: Optional[int] = Field(None)
    max_length: Optional[int] = Field(None)
    min_items: Optional[int] = Field(None)
    max_items: Optional[int] = Field(None)
    multiple_of: Optional[int] = Field(None)
    regex: Optional[str] = Field(None)
    extra: dict = Field(default_factory=dict)


def get_pait_info_from_grpc_desc(desc: str) -> MessagePaitModel:
    pait_dict: dict = {}
    for line in desc.split("\n"):
        line = line.strip()
        if not line.startswith("pait:"):
            continue
        line = line.replace("pait:", "")
        pait_dict.update(json.loads(line))
    return MessagePaitModel(**pait_dict)


def grpc_timestamp_int_handler(cls: Any, v: int) -> Timestamp:
    t: Timestamp = Timestamp()

    if v:
        t.FromDatetime(datetime.datetime.fromtimestamp(v))
    return t


def _parse_msg_to_pydantic_model(
    descriptor: Descriptor,
    grpc_timestamp_handler_tuple: GRPC_TIMESTAMP_HANDLER_TUPLE_T,
    field_doc_dict: Dict[str, str],
    default_field: Type[FieldInfo] = FieldInfo,
    request_param_field_dict: Optional[Dict[str, Union[Type[FieldInfo], Depends]]] = None,
) -> Type[BaseModel]:
    request_param_field_dict = request_param_field_dict or {}
    annotation_dict: Dict[str, Tuple[Type, Any]] = {}
    validators: Dict[str, classmethod] = {}
    timestamp_handler_field_silt: List[str] = []
    timestamp_type, _grpc_timestamp_handler = grpc_timestamp_handler_tuple

    for column in descriptor.fields:
        field: Union[Type[FieldInfo], Depends] = request_param_field_dict.get(column.name, default_field)

        type_: Any = type_dict.get(column.type, None)
        name: str = column.name
        default: Any = Undefined
        default_factory: Optional[NoArgAnyCallable] = None

        if column.type == FieldDescriptor.TYPE_MESSAGE:
            if column.message_type.name == "Timestamp":
                # support google.protobuf.Timestamp
                type_ = timestamp_type
                timestamp_handler_field_silt.append(column.name)
            elif column.message_type.name.endswith("Entry"):
                # support google.protobuf.MapEntry
                key, value = column.message_type.fields
                key_type: Any = (
                    type_dict[key.type]
                    if not key.message_type
                    else _parse_msg_to_pydantic_model(key.message_type, grpc_timestamp_handler_tuple, {})
                )
                value_type: Any = (
                    type_dict[value.type]
                    if not value.message_type
                    else _parse_msg_to_pydantic_model(value.message_type, grpc_timestamp_handler_tuple, {})
                )
                type_ = Dict[key_type, value_type]
            elif column.message_type.name == "Struct":
                # support google.protobuf.Struct
                type_ = Dict[str, Any]
            else:
                # support google.protobuf.Message
                type_ = _parse_msg_to_pydantic_model(column.message_type, grpc_timestamp_handler_tuple, {})
        elif column.type == FieldDescriptor.TYPE_ENUM:
            # support google.protobuf.Enum
            type_ = IntEnum(column.enum_type.name, {v.name: v.number for v in column.enum_type.values})  # type: ignore
            default = 0
        else:
            if column.label == FieldDescriptor.LABEL_REQUIRED:
                default = Undefined
            elif column.label == FieldDescriptor.LABEL_REPEATED:
                type_ = List[type_]  # type: ignore
                default_factory = list
            else:
                default = column.default_value

        if isinstance(field, Depends):
            use_field: Any = field
        else:
            if name in field_doc_dict:
                msg_pait_model: MessagePaitModel = get_pait_info_from_grpc_desc(field_doc_dict[name])
                field_param_dict: dict = msg_pait_model.dict()
                if not field_param_dict.pop("enable"):
                    continue
                if msg_pait_model.miss_default is not True:
                    field_param_dict["default"] = default
                field_param_dict["default_factory"] = default_factory
                use_field = field(**field_param_dict)
            else:
                use_field = field(default=default, default_factory=default_factory)
        annotation_dict[name] = (type_, use_field)

    if timestamp_handler_field_silt and _grpc_timestamp_handler:
        validators["timestamp_validator"] = validator(
            *timestamp_handler_field_silt,
            allow_reuse=True,
            check_fields=True,
            always=True,
        )(_grpc_timestamp_handler)
    return create_pydantic_model(annotation_dict, class_name=descriptor.name, pydantic_validators=validators or None)


def parse_msg_to_pydantic_model(
    msg: Union[Type[Message], Descriptor],
    default_field: Type[FieldInfo] = FieldInfo,
    request_param_field_dict: Optional[Dict[str, Union[Type[FieldInfo], Depends]]] = None,
    grpc_timestamp_handler_tuple: Optional[GRPC_TIMESTAMP_HANDLER_TUPLE_T] = None,
) -> Type[BaseModel]:
    """
    Parse a message to a pydantic model
    :param msg: grpc Message or descriptor
    :param default_field: gen pydantic_model default Field,
        apply only to the outermost pydantic model
    :param request_param_field_dict: The generated pydantic model corresponds to the field of the attribute,
        apply only to the outermost pydantic model
    :param grpc_timestamp_handler_tuple:
    """
    return _parse_msg_to_pydantic_model(
        msg if isinstance(msg, Descriptor) else msg.DESCRIPTOR,
        grpc_timestamp_handler_tuple or (str, None),
        getattr(msg, "__field_doc_dict__", {}),
        default_field=default_field,
        request_param_field_dict=request_param_field_dict,
    )



from typing import Any

from tornado.web import RequestHandler

from pait.plugin.auto_complete_json_resp import AutoCompleteJsonRespPlugin as _AutoCompleteJsonRespPlugin

from .base import JsonProtocol

__all__ = ["AsyncAutoCompleteJsonRespPlugin", "AutoCompleteJsonRespPlugin"]


class AutoCompleteJsonRespPlugin(JsonProtocol, _AutoCompleteJsonRespPlugin):
    async def __call__(self, *args: Any, **kwargs: Any) -> Any:
        response: Any = await super(AutoCompleteJsonRespPlugin, self).__call__(*args, **kwargs)
        tornado_handle: RequestHandler = args[0]
        return self.gen_response(tornado_handle, response)


class AsyncAutoCompleteJsonRespPlugin(AutoCompleteJsonRespPlugin):
    """"""



from pait.app.flask.plugin.check_json_resp import CheckJsonRespPlugin
from pait.app.flask.plugin.mock_response import MockPlugin
from pait.plugin.at_most_one_of import AtMostOneOfPlugin
from pait.plugin.required import RequiredPlugin

__all__ = ["AtMostOneOfPlugin", "RequiredPlugin", "CheckJsonRespPlugin", "MockPlugin"]



from ._app_helper import AppHelper
from ._doc_route import AddDocRoute, add_doc_route
from ._load_app import load_app
from ._pait import Pait, pait
from ._test_helper import StarletteTestHelper, TestHelper



from contextvars import ContextVar, Token
from typing import Any, Dict, Optional

from pydantic.fields import Undefined

TEMPLATE_CONTEXT: ContextVar[Dict[str, Any]] = ContextVar("template_context", default={})

__all__ = ["TemplateContext", "TemplateVar"]


class TemplateContext(object):
    def __init__(self, template_dict: Dict[str, Any]) -> None:
        self._token: Optional[Token] = None
        self._template_dict: Dict[str, Any] = template_dict

    def __enter__(self) -> "TemplateContext":
        self._token = TEMPLATE_CONTEXT.set(self._template_dict)
        return self

    def __exit__(self, exc_type: Any, exc_val: Any, exc_tb: Any) -> None:
        if self._token:
            TEMPLATE_CONTEXT.reset(self._token)
            self._token = None


class TemplateVar(object):
    def __init__(self, name: str, default_value: Any = Undefined) -> None:
        self._name: str = name
        self._default_value: Any = default_value

    def __call__(self) -> Any:
        return self.get_value_from_template_context()

    def get_value_from_template_context(self) -> Any:
        return TEMPLATE_CONTEXT.get().get(self._name, self._default_value)



import asyncio
import inspect
import json
from abc import ABCMeta
from sys import modules
from typing import Any, Callable, Dict, List, Optional, Tuple, Type, Union

import grpc
from pydantic import BaseModel, Field

from pait.core import Pait
from pait.field import BaseField, Body, Depends
from pait.model.response import PaitJsonResponseModel
from pait.model.tag import Tag
from pait.util.grpc_inspect.message_to_pydantic import GRPC_TIMESTAMP_HANDLER_TUPLE_T, parse_msg_to_pydantic_model
from pait.util.grpc_inspect.stub import GrpcModel, ParseStub
from pait.util.grpc_inspect.types import Message, MessageToDict


class GrpcPaitModel(BaseModel):
    name: str = Field("")
    tag: List[Tuple[str, str]] = Field(default_factory=list)
    group: str = Field("")
    desc: str = Field("")
    summary: str = Field("")
    url: str = Field("")
    enable: bool = Field(True)


def get_pait_info_from_grpc_desc(grpc_model: GrpcModel) -> GrpcPaitModel:
    pait_dict: dict = {}
    for line in grpc_model.service_desc.split("\n") + grpc_model.desc.split("\n"):
        line = line.strip()
        if not line.startswith("pait: {"):
            continue
        line = line.replace("pait:", "")
        pait_dict.update(json.loads(line))
    return GrpcPaitModel(**pait_dict)


class BaseGrpcGatewayRoute(object):
    pait: Pait
    make_response: Callable
    channel: Union[grpc.Channel, grpc.aio.Channel]

    _grpc_tag: Tag = Tag("grpc", desc="grpc route")

    def __init__(
        self,
        app: Any,
        *stub_list: Any,
        parse_msg_desc: Optional[str] = None,
        prefix: str = "",
        title: str = "",
        msg_to_dict: Callable = MessageToDict,
        parse_dict: Optional[Callable] = None,
        pait: Optional[Pait] = None,
        make_response: Optional[Callable] = None,
        url_handler: Callable[[str], str] = lambda x: x.replace(".", "-"),
        request_param_field_dict: Optional[Dict[str, Union[Type[BaseField], Depends]]] = None,
        grpc_timestamp_handler_tuple: Optional[GRPC_TIMESTAMP_HANDLER_TUPLE_T] = None,
    ):
        self.prefix: str = prefix
        self.title: str = title
        self.parse_stub_list: List[ParseStub] = [ParseStub(i, parse_msg_desc=parse_msg_desc) for i in stub_list]
        self.stub_list: Tuple[Any, ...] = stub_list
        self.msg_to_dict: Callable = msg_to_dict
        self.parse_dict: Optional[Callable] = parse_dict

        self.url_handler: Callable[[str], str] = url_handler
        self._request_param_field_dict: Optional[Dict[str, Union[Type[BaseField], Depends]]] = request_param_field_dict
        self._grpc_timestamp_handler_tuple: Optional[GRPC_TIMESTAMP_HANDLER_TUPLE_T] = grpc_timestamp_handler_tuple
        self._pait: Pait = pait or self.pait
        self._make_response: Callable = make_response or self.make_response
        self._is_gen: bool = False
        self._tag_dict: Dict[str, Tag] = {}
        self.method_func_dict: Dict[str, Callable] = {}

        self._add_route(app)

    def _gen_request_pydantic_class_from_grpc_model(self, grpc_model: GrpcModel) -> Type[BaseModel]:
        return parse_msg_to_pydantic_model(
            grpc_model.request,
            default_field=Body,
            request_param_field_dict=self._request_param_field_dict,
            grpc_timestamp_handler_tuple=self._grpc_timestamp_handler_tuple,
        )

    def _gen_pait_from_grpc_method(
        self, method_name: str, grpc_model: GrpcModel, grpc_pait_model: GrpcPaitModel
    ) -> Pait:
        tag_tuple: Tuple[Tag, ...] = (self._grpc_tag,)
        for tag, desc in grpc_pait_model.tag + [("grpc" + "-" + method_name.split("/")[1].split(".")[0], "")]:
            if tag in self._tag_dict:
                pait_tag: Tag = self._tag_dict[tag]
            else:
                pait_tag = Tag(tag, desc)
                self._tag_dict[tag] = pait_tag
            tag_tuple += (pait_tag,)

        class CustomerJsonResponseModel(PaitJsonResponseModel):
            name: str = grpc_model.response.DESCRIPTOR.name
            response_data: Type[BaseModel] = parse_msg_to_pydantic_model(grpc_model.response)

        return self._pait.create_sub_pait(
            name=grpc_pait_model.name,
            group=grpc_pait_model.group or method_name.split("/")[1],
            tag=tag_tuple,
            desc=grpc_pait_model.desc,
            summary=grpc_pait_model.summary,
            response_model_list=[CustomerJsonResponseModel],
        )

    def get_grpc_func(self, method_name: str) -> Callable:
        func: Optional[Callable] = self.method_func_dict.get(method_name, None)
        if not func:
            raise RuntimeError(f"{method_name}'s func is not found, Please call init_channel to register the channel")
        return func

    def get_msg_from_dict(self, msg: Type[Message], request_dict: dict) -> Message:
        if self.parse_dict:
            request_msg: Message = self.parse_dict(request_dict, msg)
        else:
            request_msg = msg(**request_dict)  # type: ignore
        return request_msg

    def get_dict_from_msg(self, msg: Message) -> dict:
        return self.msg_to_dict(msg)

    def gen_route(self, method: str, grpc_model: GrpcModel, request_pydantic_model_class: Type[BaseModel]) -> Callable:
        raise NotImplementedError()

    def _gen_route_func(self, method_name: str, grpc_model: GrpcModel) -> Tuple[Optional[Callable], GrpcPaitModel]:
        grpc_pait_model: GrpcPaitModel = get_pait_info_from_grpc_desc(grpc_model)
        if grpc_pait_model.enable is False:
            return None, grpc_pait_model
        if not grpc_pait_model.url:
            grpc_pait_model.url = method_name

        request_pydantic_model_class: Type[BaseModel] = self._gen_request_pydantic_class_from_grpc_model(grpc_model)
        pait: Pait = self._gen_pait_from_grpc_method(method_name, grpc_model, grpc_pait_model)

        _route = self.gen_route(method_name, grpc_model, request_pydantic_model_class)
        # request_pydantic_model_class is not generated by this module,
        # so you need to inject request_pydantic_model_class into this module.
        sig: inspect.Signature = inspect.signature(_route)
        for key in sig.parameters:
            if sig.parameters[key].annotation == sig.empty:
                continue

        modules[_route.__module__].__dict__["request_pydantic_model_class"] = request_pydantic_model_class
        # change route func name and qualname
        _route.__name__ = self.title + method_name.replace(".", "_")
        _route.__qualname__ = _route.__qualname__.replace("._route", "." + _route.__name__)

        _route = pait(feature_code=grpc_model.method)(_route)
        return _route, grpc_pait_model

    def _add_route(self, app: Any) -> Any:  # type: ignore
        raise NotImplementedError()

    def reinit_channel(
        self, channel: Union[grpc.Channel, grpc.aio.Channel]
    ) -> Union[grpc.Channel, grpc.aio.Channel, None]:
        old_channel: Union[grpc.Channel, grpc.aio.Channel, None] = getattr(self, "channel", None)
        self.init_channel(channel)
        return old_channel

    def _init_channel(self, channel: Union[grpc.Channel, grpc.aio.Channel]) -> None:
        self.channel: Union[grpc.Channel, grpc.aio.Channel] = channel
        for stub_class in self.stub_list:
            stub = stub_class(channel)
            for func in stub.__dict__.values():
                method = func._method  # type: ignore
                if isinstance(method, bytes):
                    method = method.decode()
                self.method_func_dict[method] = func

    def init_channel(self, channel: Union[grpc.Channel, grpc.aio.Channel]) -> None:
        raise NotImplementedError()


class GrpcGatewayRoute(BaseGrpcGatewayRoute, metaclass=ABCMeta):
    def gen_route(
        self, method_name: str, grpc_model: GrpcModel, request_pydantic_model_class: Type[BaseModel]
    ) -> Callable:
        def _route(request_pydantic_model: request_pydantic_model_class) -> Any:  # type: ignore
            func: Callable = self.get_grpc_func(method_name)
            request_dict: dict = request_pydantic_model.dict()  # type: ignore
            request_msg: Message = self.get_msg_from_dict(grpc_model.request, request_dict)
            grpc_msg: Message = func(request_msg)
            return self._make_response(self.get_dict_from_msg(grpc_msg))

        return _route

    def init_channel(self, channel: grpc.Channel) -> None:
        self._init_channel(channel)


class AsyncGrpcGatewayRoute(BaseGrpcGatewayRoute, metaclass=ABCMeta):
    def gen_route(
        self, method_name: str, grpc_model: GrpcModel, request_pydantic_model_class: Type[BaseModel]
    ) -> Callable:
        async def _route(request_pydantic_model: request_pydantic_model_class) -> Any:  # type: ignore
            func: Callable = self.get_grpc_func(method_name)
            request_dict: dict = request_pydantic_model.dict()  # type: ignore
            request_msg: Message = self.get_msg_from_dict(grpc_model.request, request_dict)
            loop: asyncio.AbstractEventLoop = asyncio.get_event_loop()
            if loop != func._loop:  # type: ignore
                raise RuntimeError(
                    "Loop is not same, "
                    "the grpc channel must be initialized after the event loop of the api server is initialized"
                )
            else:
                grpc_msg: Message = await func(request_msg)
            return self._make_response(self.get_dict_from_msg(grpc_msg))

        return _route

    def init_channel(self, channel: grpc.aio.Channel) -> None:
        self._init_channel(channel)



from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from pait.model.config import Config
    from pait.model.core import PaitCoreModel


__all__ = ["sync_config_data_to_pait_core_model"]


def sync_config_data_to_pait_core_model(config: "Config", pait_core_model: "PaitCoreModel", **kwargs: Any) -> None:
    if not pait_core_model.author:
        pait_core_model.author = config.author
    if not pait_core_model.status:
        pait_core_model.status = config.status

    if config.apply_func_list:
        for apply_func in config.apply_func_list:
            apply_func(pait_core_model)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pait',
 'pait.api_doc',
 'pait.api_doc.html',
 'pait.app',
 'pait.app.base',
 'pait.app.flask',
 'pait.app.flask.plugin',
 'pait.app.sanic',
 'pait.app.sanic.plugin',
 'pait.app.starlette',
 'pait.app.starlette.plugin',
 'pait.app.tornado',
 'pait.app.tornado.plugin',
 'pait.extra',
 'pait.model',
 'pait.plugin',
 'pait.util',
 'pait.util.grpc_inspect']

package_data = \
{'': ['*']}

install_requires = \
['pydantic>=1.7.3,<2.0.0', 'typing-extensions>=4.1.1,<5.0.0']

extras_require = \
{'redis': ['redis>=4.2.2,<5.0.0']}

setup_kwargs = {
    'name': 'pait',
    'version': '0.7.8.1',
    'description': 'Pait is a Python api tool. Pait enables your Python web framework to have type checking, parameter type conversion, interface document generation and can display your documents through Redoc or Swagger (power by inspect, pydantic)',
    'long_description': '![](https://cdn.jsdelivr.net/gh/so1n/so1n_blog_photo@master/blog_photo/1652600629491%E6%9C%AA%E5%91%BD%E5%90%8D.jpg)\n<p align="center">\n    <em>Python Api Tools, fast to code</em>\n</p>\n\n---\n**Documentation**: [https://so1n.me/pait/](https://so1n.me/pait/)\n\n**中文文档**: [https://so1n.me/pait-zh-doc/](https://so1n.me/pait-zh-doc/)\n\n---\n\n# pait\nPait is an api tool that can be used in any python web framework (currently only `flask`, `starlette`, `sanic`, `tornado` are supported, other frameworks will be supported once Pait is stable).\n\n> Note:\n>\n> mypy check 100%\n>\n> test coverage 95%+ (exclude api_doc)\n>\n> python version >= 3.7 (support postponed annotations)\n>\n> The following code does not specify, all default to use the `starlette` framework.\n\n# Feature\n - [x] Parameter checksum automatic conversion (parameter check depends on `Pydantic`)\n - [x] Parameter dependency verification\n - [x] Automatically generate openapi files\n - [x] Swagger, Redoc route\n - [x] gRPC Gateway route\n - [x] TestClient support, support response result verification\n - [x] Support for plugin extensions, such as the Mock plugin\n\n# Installation\n```Bash\npip install pait\n```\n# Simple Example\n```python\nfrom typing import Type\nimport uvicorn  # type: ignore\nfrom starlette.applications import Starlette\nfrom starlette.responses import JSONResponse\nfrom starlette.routing import Route\n\nfrom pait.app.starlette import pait, add_doc_route\nfrom pait.field import Body\nfrom pait.model.response import PaitResponseModel\nfrom pydantic import BaseModel, Field\n\n\nclass DemoResponseModel(PaitResponseModel):\n    class ResponseModel(BaseModel):\n        uid: int = Field()\n        user_name: str = Field()\n\n    description: str = "demo response"\n    response_data: Type[BaseModel] = ResponseModel\n\n\n@pait(response_model_list=[DemoResponseModel])\nasync def demo_post(\n    uid: int = Body.i(description="user id", gt=10, lt=1000),\n    user_name: str = Body.i(description="user name", min_length=2, max_length=4)\n) -> JSONResponse:\n    return JSONResponse({\'uid\': uid, \'user_name\': user_name})\n\n\napp = Starlette(routes=[Route(\'/api\', demo_post, methods=[\'POST\'])])\nadd_doc_route(app)\nuvicorn.run(app)\n```\n\n# How to used in other web framework?\nIf the web framework is not supported, which you are using.\n\nCan be modified sync web framework according to [pait.app.flask](https://github.com/so1n/pait/blob/master/pait/app/flask.py)\n\nCan be modified async web framework according to [pait.app.starlette](https://github.com/so1n/pait/blob/master/pait/app/starlette.py)\n# IDE Support\nWhile pydantic will work well with any IDE out of the box.\n- [PyCharm plugin](https://pydantic-docs.helpmanual.io/pycharm_plugin/)\n- [Mypy plugin](https://pydantic-docs.helpmanual.io/mypy_plugin/)\n\n# Full example\nFor more complete examples, please refer to [example](https://github.com/so1n/pait/tree/master/example)\n',
    'author': 'So1n',
    'author_email': 'so1n897046026@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/so1n/pait',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



from typing import Any, List

from pait.exceptions import CheckValueError
from pait.plugin.base import PluginManager, PluginProtocol
from pait.util import gen_tip_exc

__all__ = ["AsyncAtMostOneOfPlugin", "AtMostOneOfPlugin"]


class AtMostOneOfPlugin(PluginProtocol):
    """
    Check whether each group of parameters appear at the same time
    """

    at_most_one_of_list: List[List[str]]

    is_pre_core: bool = False

    def check_param(self, **kwargs: Any) -> None:
        try:
            for at_most_one_of in self.at_most_one_of_list:
                if len([i for i in at_most_one_of if kwargs.get(i, None) is not None]) > 1:
                    raise CheckValueError(f"requires at most one of param {' or '.join(at_most_one_of)}")
        except Exception as e:
            raise e from gen_tip_exc(self.pait_core_model.func, e)

    @classmethod
    def build(cls, *, at_most_one_of_list: List[List[str]]) -> "PluginManager":  # type: ignore
        return super().build(at_most_one_of_list=at_most_one_of_list)

    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        self.check_param(**kwargs)
        return self.call_next(*args, **kwargs)


class AsyncAtMostOneOfPlugin(AtMostOneOfPlugin):
    """"""
